from cohdl import std, Variable, BitVector, Null

from . import reg


class Interface:
    def __init__(self):
        self._ctx: std.SequentialContext

    def read_handler(self):
        ...

    def write_handler(self):
        ...


def register_interface(interface: Interface, reg_device: reg.RegisterDevice):
    flat = reg_device._flatten_()

    for elem in flat:
        if type(elem)._impl_ is not reg.RegisterObject._impl_:
            elem._impl_(interface._ctx)

        if type(elem)._impl_sequential_ is not reg.RegisterObject._impl_sequential_:
            interface._ctx(elem._impl_sequential_)

        if type(elem)._impl_concurrent_ is not reg.RegisterObject._impl_concurrent_:
            std.concurrent(elem._impl_concurrent_)

    async def on_read(addr):
        result = Variable[BitVector[32]](Null)

        for elem in flat:
            if elem._readable_:
                if elem._addr_offset_ == addr:
                    result @= await std.as_awaitable(elem.on_read_base, addr)
                    break

        return result

    async def on_write(addr, data, mask):
        for elem in flat:
            if elem._writable_:
                if elem._addr_offset_ == addr:
                    await std.as_awaitable(elem.on_write_base, addr, data, mask)



from __future__ import annotations

from typing import Tuple


from cohdl._core._type_qualifier import TypeQualifier, Signal, Variable, Temporary
from cohdl._core._primitive_type import is_primitive

from cohdl._core._intrinsic import _BitSignalEvent, _BitSignalEventGroup
from cohdl._core._bit import Bit
from cohdl._core._boolean import _Boolean, _BooleanLiteral, _NullFullType
from cohdl._core._bit_vector import BitVector
from cohdl._core._signed import Signed
from cohdl._core._unsigned import Unsigned


class _Redirect:
    def __init__(self, target, source):
        self.target = target

        if isinstance(source, TypeQualifier):
            try:
                # try to construct target type from source to check,
                # that they are compatible
                target.type(TypeQualifier.decay(source))
            except Exception as err:
                err.add_note(
                    f"at least one possible source of this expression is not compatible with the target (source={source})"
                )
                raise

            self.source = source
        else:
            self.source = target.type(source)

    def print(self):
        print(f"{self.target} <<= {self.source}")


class _ValueBranchHook:
    def __init__(self, redirects: list[_Redirect] | None = None, *, name=None):
        self.redirects = [] if redirects is None else redirects
        self.name = name

    def add_redirect(self, redirect: _Redirect):
        self.redirects.append(redirect)

    def has_redirect(self):
        return len(self.redirects) != 0

    def print(self):
        for red in self.redirects:
            red.print()


class _ValueBranch:
    def __init__(self, hook: _ValueBranchHook, obj):
        self.obj = obj
        self.hook = hook

    def _type(self):
        return _ValueBranch(self.hook, type(self.obj))

    def _id(self):
        return _ValueBranch(self.hook, id(self.obj))

    def __getitem__(self, args):
        return _ValueBranch(self.hook, self.obj.__getitem__(args))

    def _getattr(self, name):
        return _ValueBranch(self.hook, ObjTraits.getattr(self.obj, name))

    def _redirect(self, target):
        assert isinstance(target, TypeQualifier)

        if isinstance(self.obj, _MergedBranch):
            source_obj = Temporary[target.type]()
            self.obj._redirect_values(source_obj)
        else:
            source_obj = self.obj

        self.hook.add_redirect(_Redirect(target, source_obj))


def _try_join(options):
    result_type = None

    # find most constrained primitive type compatible with all options
    for option in options:
        option = TypeQualifier.decay(option)
        option_type = type(option)

        if result_type is None:
            if is_primitive(option):
                result_type = type(option)
            elif isinstance(option, bool):
                result_type = _Boolean
        else:
            if issubclass(result_type, Bit):
                if option_type not in (
                    Bit,
                    bool,
                    _Boolean,
                    _BooleanLiteral,
                    _NullFullType,
                ):
                    # incompatible branches, return early
                    return None
            elif issubclass(result_type, _Boolean):
                if option_type is Bit:
                    result_type = Bit
                elif option_type not in (
                    bool,
                    _Boolean,
                    _BooleanLiteral,
                    _NullFullType,
                ):
                    # incompatible branches, return early
                    return None
            elif issubclass(result_type, BitVector) and issubclass(
                option_type, BitVector
            ):
                if issubclass(result_type, Signed) and issubclass(option_type, Signed):
                    result_type = (
                        result_type
                        if result_type.width >= option_type.width
                        else option_type
                    )
                elif issubclass(result_type, Unsigned) and issubclass(
                    option_type, Unsigned
                ):
                    result_type = (
                        result_type
                        if result_type.width >= option_type.width
                        else option_type
                    )
                else:
                    if result_type.width != option_type.width:
                        # incompatible branches, return early
                        return None

                    result_type = BitVector[result_type.width]

    if result_type is None:
        return None

    for option in options:
        option = TypeQualifier.decay(option)
        try:
            result_type(option)
        except Exception:
            return None

    return result_type


class _MergedBranch:
    def __new__(cls, branches: list[_ValueBranch]):
        if len(branches) == 0:
            return None

        if len(branches) == 1:
            return branches[0].obj

        first, *rest = branches
        first_obj = first.obj

        if all(first.obj is branch.obj for branch in rest):
            return first_obj

        result_type = _try_join([branch.obj for branch in branches])

        if result_type is not None:
            result = Temporary[result_type]()

            for branch in branches:
                branch._redirect(result)

            return result

        return object.__new__(cls)

    def __init__(self, branches: list[_ValueBranch]):
        # avoid double initialization if __new__ returns an existing
        # _MergedBranch object (possible when branches contains a single _MergedBranch object)
        if not hasattr(self, "_cohdl_init_complete"):
            for branch in branches:
                assert isinstance(
                    branch, _ValueBranch
                ), "internal error: argument of MergedBranch is not a _ValueBranch"
                assert branch.obj is not self, "internal error: recursive merged branch"

            self.branches = branches
            self._cohdl_init_complete = True

    def isinstance(self, types):
        branch_results = [isinstance(branch.obj, types) for branch in self.branches]

        all_results = all(branch_results)
        any_result = any(branch_results)

        assert (
            all_results or not any_result
        ), "the result of isinstance must be the same for all possible input objects"
        return all_results

    def issubclass(self, types):
        branch_results = [issubclass(branch.obj, types) for branch in self.branches]

        all_results = all(branch_results)
        any_result = any(branch_results)

        assert (
            all_results or not any_result
        ), "the result of issubclass must be the same for all possible input objects"
        return all_results

    def type(self):
        return _MergedBranch([branch._type() for branch in self.branches])

    def id(self):
        return _MergedBranch([branch._id() for branch in self.branches])

    def _getitem(self, *args):
        return _MergedBranch([branch.__getitem__(*args) for branch in self.branches])

    def _hasattr(self, name):
        return all(ObjTraits.hasattr(branch.obj, name) for branch in self.branches)

    def _getattr(self, name):
        return _MergedBranch([branch._getattr(name) for branch in self.branches])

    def _redirect_values(self, target):
        for branch in self.branches:
            branch._redirect(target)


class ObjTraits:
    @staticmethod
    def init_active(obj) -> bool:
        if isinstance(obj, _MergedBranch):
            return False
        return hasattr(obj, "_cohdl_init_active")

    @staticmethod
    def hasattr(obj, name: str):
        if isinstance(obj, _MergedBranch):
            return obj._hasattr(name)

        # don't use hasattr directly to avoid problems
        # with descriptors

        if isinstance(obj, int):
            return hasattr(obj, name)

        return (
            (hasattr(obj, "__dict__") and name in vars(obj))
            or hasattr(type(obj), name)
            or hasattr(obj, name)
        )

    @staticmethod
    def gettype(obj):
        if isinstance(obj, _MergedBranch):
            return obj.type()
        return type(obj)

    @staticmethod
    def getattr(obj, name: str):
        if isinstance(obj, _MergedBranch):
            return obj._getattr(name)
        return getattr(obj, name)

    @staticmethod
    def isinstance(obj, class_or_tuple: type | Tuple) -> bool:
        assert not isinstance(obj, _MergedBranch)
        if isinstance(obj, _MergedBranch):
            return obj.is_instance(class_or_tuple)
        return isinstance(obj, class_or_tuple)

    @staticmethod
    def get(obj):
        assert not isinstance(obj, _MergedBranch)
        if isinstance(obj, _MergedBranch):
            return obj.get()
        return obj

    @staticmethod
    def runtime_variable(obj) -> bool:
        assert not isinstance(obj, _MergedBranch)
        return (ObjTraits.isinstance(obj, (Signal, Variable, Temporary))) or isinstance(
            obj, (_BitSignalEvent, _BitSignalEventGroup)
        )



from ._compile import VhdlCompiler
from ._assignable_type import AssignableType

from ._context import (
    concurrent,
    sequential,
    block,
    Clock,
    ClockEdge,
    Reset,
    Frequency,
    Period,
    Duration,
    Context,
    SequentialContext,
    concurrent_assign,
    concurrent_call,
    concurrent_eval,
    Executor,
    ExecutorMode,
)


ms = Duration.milliseconds
us = Duration.microseconds
ns = Duration.nanoseconds
ps = Duration.picoseconds

kHz = Frequency.kilohertz
MHz = Frequency.megahertz
GHz = Frequency.gigahertz

from . import axi

from ._core_utility import (
    nop,
    comment,
    as_pyeval,
    fail,
    identity,
    regenerate_defaults,
    Ref,
    Value,
    Signal,
    Variable,
    Temporary,
    Nonlocal,
    NoresetSignal,
    NoresetVariable,
    base_type,
    instance_check,
    subclass_check,
    iscouroutinefunction,
    as_awaitable,
    zeros,
    ones,
    width,
    one_hot,
    reverse_bits,
    is_qualified,
    const_cond,
    check_type,
    select,
    choose_first,
    cond,
    to_bits,
    from_bits,
    count_bits,
    check_return,
    binary_fold,
    batched_fold,
    concat,
    stretch,
    leftpad,
    rightpad,
    pad,
    apply_mask,
    Mask,
    as_bitvector,
    rol,
    ror,
    lshift_fill,
    rshift_fill,
    batched,
    select_batch,
)

from . import utility
from .utility import (
    add_entity_port,
    Serialized,
    as_readable_vector,
    as_writeable_vector,
    Array,
    tick,
    wait_for,
    wait_forever,
    Waiter,
    stringify,
    DelayLine,
    delayed,
    debounce,
    max_int,
    int_log_2,
    ceil_log_2,
    is_pow_two,
    InShiftRegister,
    OutShiftRegister,
    continuous_counter,
    ToggleSignal,
    ClockDivider,
    SyncFlag,
    Mailbox,
    Fifo,
)

from ._fixed import SFixed, UFixed, FixedOverflowStyle, FixedRoundStyle
from ._prefix import prefix, name, NamedQualifier
from . import spi
from . import bitfield
from . import _crc as crc
from . import uart

from ._template import Template, template_arg
from ._record import Record
from .enum import Enum, FlagEnum

# from . import experimental as exp



from __future__ import annotations

import cohdl

Signal = cohdl.Signal
Bit = cohdl.Bit


class Channel:
    def __init__(self, valid, ready, **kwargs):
        self.valid = valid
        self.ready = ready

        self._signals = kwargs

    async def send(self, **kwargs):
        for name, value in kwargs.items():
            self._signals[name] <<= value

        self.valid <<= True

        await self.ready

        self.valid <<= False

        for value in self._signals.values():
            value <<= cohdl.Null

    async def receive(self):
        self.ready <<= True

        await self.valid

        self.ready <<= False

        return self._signals



from __future__ import annotations

import enum

import typing
from typing import overload

from cohdl.utility import Span
from cohdl._core._unsigned import Unsigned
from cohdl._core._signed import Signed

from cohdl._core._bit import Bit, BitState
from cohdl._core._integer import Integer
from cohdl._core._boolean import _NullFullType
from cohdl._core._primitive_type import _PrimitiveType

class InvalidOverload: ...

class BitOrder(enum.Enum):
    DOWNTO = enum.auto()
    UPTO = enum.auto()

class BitVector(_PrimitiveType):
    def __class_getitem__(cls, offset) -> type[BitVector]: ...
    @property
    def width(self) -> int: ...
    @property
    def order(self) -> BitVector: ...
    def __init__(
        self,
        val: None | BitVector | str | _NullFullType | Span[Bit] = None,
    ) -> None: ...
    def __iter__(self) -> list[Bit]: ...
    def copy(self) -> BitVector: ...
    @overload
    def right(self, width=None, rest=None) -> Bit: ...
    @overload
    def right(self, width: int | None, rest: int | None) -> BitVector: ...
    @overload
    def left(self, width=None, rest=None) -> Bit: ...
    @overload
    def left(self, width: int | None, rest: int | None) -> BitVector: ...
    @overload
    def lsb(self, width=None, rest=None) -> Bit: ...
    @overload
    def lsb(self, width: int | None, rest: int | None) -> BitVector: ...
    @overload
    def msb(self, width=None, rest=None) -> Bit: ...
    @overload
    def msb(self, width: int | None, rest: int | None) -> BitVector: ...
    def __bool__(self) -> bool: ...
    def __invert__(self) -> BitVector: ...
    def __inv__(self) -> BitVector: ...
    @overload
    def __and__(self, other: Signed | Unsigned) -> InvalidOverload: ...
    @overload
    def __and__(self, other: BitVector) -> BitVector: ...
    @overload
    def __or__(self, other: Signed | Unsigned) -> InvalidOverload: ...
    @overload
    def __or__(self, other: BitVector) -> BitVector: ...
    @overload
    def __xor__(self, other: Signed | Unsigned) -> InvalidOverload: ...
    @overload
    def __xor__(self, other: BitVector) -> BitVector: ...
    def __eq__(self, other: BitVector) -> bool: ...
    def __ne__(self, other: BitVector) -> bool: ...
    def __matmul__(self, rhs: Bit | BitVector) -> BitVector: ...
    def __rmatmul__(self, lhs: Bit) -> BitVector: ...
    def __len__(self) -> int: ...
    @overload
    def __getitem__(self, key: int | Integer) -> Bit: ...
    @overload
    def __getitem__(self, key: slice) -> BitVector: ...
    @overload
    def __getitem__(self, key: list | tuple) -> BitVector: ...
    @overload
    def __setitem__(self, key: int | Integer, value: Bit): ...
    @overload
    def __setitem__(self, key: slice, value: BitVector): ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    @property
    def unsigned(self: BitVector) -> Unsigned: ...
    @unsigned.setter
    def unsigned(self: BitVector, value): ...
    @property
    def signed(self) -> Signed: ...
    @signed.setter
    def signed(self, value: BitVector): ...
    @property
    def bitvector(self) -> BitVector: ...
    @bitvector.setter
    def bitvector(self, value: BitVector): ...



from __future__ import annotations
from dataclasses import dataclass
from typing import Any

import enum

from cohdl.utility import IdMap

#
#
# intrinsic decorators
#
#


class _IntrinsicReplacement:
    class _Assignment:
        def __init__(self, nr_target, nr_source):
            self.nr_target = nr_target
            self.nr_source = nr_source

    def __init__(self, fn, is_special_case, *, assignment_spec=None, evaluate=False):
        self.fn = fn

        if evaluate:
            assert (
                is_special_case
            ), "internal error: cannot evaluate _IntrinsicReplacement unless it is special cased"

        self.is_special_case = is_special_case
        self.assignment_spec = assignment_spec
        self.evaluate = evaluate


_intrinsic_functions = []
_intrinsic_replacements = IdMap[Any, _IntrinsicReplacement]()


def _intrinsic(fn):
    _intrinsic_functions.append(fn)
    return fn


def _intrinsic_replacement(
    default_implementation,
    *,
    special_case=True,
    assignment_spec: tuple[int, int] | None = None,
    evaluate=False,
):
    assert (
        default_implementation in _intrinsic_functions
    ), "internal error: default implementation of intrinsic_replacement is not intrinsic"

    if assignment_spec is not None:
        assignment_spec = _IntrinsicReplacement._Assignment(*assignment_spec)

    def helper(fn):
        assert (
            default_implementation not in _intrinsic_replacements
        ), "internal error: redefinition of intrinsic_replacement"
        _intrinsic_replacements[default_implementation] = _IntrinsicReplacement(
            fn, special_case, assignment_spec=assignment_spec, evaluate=evaluate
        )
        return fn

    return helper


def _intrinsic_property(p):
    _intrinsic(p.fset)
    _intrinsic(p.fget)
    return p


def _is_intrinsic(fn):
    return fn in _intrinsic_functions


def _has_intrinsic_replacement(fn):
    return fn in _intrinsic_replacements


#
# comment
#


class _IntrinsicComment:
    def __init__(self, lines: list[str]):
        self.lines = lines


@_intrinsic
def comment(*lines: str):
    assert all(
        isinstance(line, str) for line in lines
    ), "all arguments of comment must be of type str"
    pass


@_intrinsic_replacement(comment)
def _comment_replacement(*lines: str):
    assert all(
        isinstance(line, str) for line in lines
    ), "all arguments of comment must be of type str"
    return _IntrinsicComment(lines)


#
#
# select
#
#


@_intrinsic
def select_with(arg, branches: dict, default=None):
    if arg in branches:
        return branches[arg]

    return default


@dataclass
class _SelectWith:
    arg: Any
    branches: dict
    default: Any


@_intrinsic_replacement(select_with)
def _intrinsic_select_with(arg, branches: dict, default=None):
    return _SelectWith(arg, branches, default)


#
#
#


class _CoroutineStep:
    def __init__(self, coro):
        self.coro = coro


@_intrinsic
def coroutine_step(coro) -> None:
    """
    should never be called, always is handled by parser
    """

    raise AssertionError("coroutine_step called outside synthesizable context")


@_intrinsic_replacement(coroutine_step)
def coroutine_step_replacement(coro):
    return _CoroutineStep(coro)


class _SensitivitySpec: ...


class _SensitivityList(_SensitivitySpec):
    def __init__(self, signals):
        self.signals: list = signals


class _SensitivityAll(_SensitivitySpec):
    def __init__(self): ...


class _Sensitivity:
    @staticmethod
    def list(*args):
        raise AssertionError("_Sensitivity.list called outside synthesizable context")

    @staticmethod
    def all():
        raise AssertionError("_Sensitivity.all called outside synthesizable context")


_intrinsic(_Sensitivity.all)
_intrinsic(_Sensitivity.list)


@_intrinsic_replacement(_Sensitivity.list)
def sensitifity_list_replacement(*args):
    return _SensitivityList([*args])


@_intrinsic_replacement(_Sensitivity.all)
def sensitifity_all_replacement(*args):
    return _SensitivityAll()


sensitivity = _Sensitivity

#
#
#


class _ResetContext: ...


@_intrinsic
def reset_context():
    raise AssertionError("reset_context called outside synthesizable context")


@_intrinsic_replacement(reset_context)
def reset_context_replacement():
    return _ResetContext()


class _ResetPushed: ...


@_intrinsic
def reset_pushed():
    raise AssertionError("reset_pushed called outside synthesizable context")


@_intrinsic_replacement(reset_pushed)
def reset_pushed_replacement():
    return _ResetPushed()


#
#
#


class _IsInstance:
    def __init__(self, obj, type):
        self.obj = obj
        self.type = type


class _IsSubclass:
    def __init__(self, cls, type):
        self.cls = cls
        self.type = type


class _Type:
    def __init__(self, obj):
        self.obj = obj


class _Id:
    def __init__(self, obj):
        self.obj = obj


_intrinsic(isinstance)
_intrinsic(issubclass)
_intrinsic(type)
_intrinsic(id)


@_intrinsic_replacement(isinstance)
def isinstance_replacement(obj, type):
    return _IsInstance(obj, type)


@_intrinsic_replacement(issubclass)
def issubclass_replacement(obj, type):
    return _IsSubclass(obj, type)


@_intrinsic_replacement(type)
def type_replacement(obj):
    return _Type(obj)


@_intrinsic_replacement(id)
def id_replacement(obj):
    return _Id(obj)


#
#
#


class _BitSignalEvent:
    class Type(enum.Enum):
        RISING = enum.auto()
        FALLING = enum.auto()
        BOTH_EDGES = enum.auto()
        HIGH = enum.auto()
        LOW = enum.auto()

    def __init__(self, sig, event_type: _BitSignalEvent.Type):
        self.sig = sig
        self.event_type = event_type

    def __and__(self, other):
        return _BitSignalEventGroup(_BitSignalEventGroup.Operation.AND, [self, other])

    def __or__(self, other):
        return _BitSignalEventGroup(_BitSignalEventGroup.Operation.OR, [self, other])


class _BitSignalEventGroup:
    class Operation(enum.Enum):
        AND = enum.auto()
        OR = enum.auto()

    def __init__(self, operation, events: list[_BitSignalEvent | _BitSignalEventGroup]):
        self.operation = operation
        self.events = events

    def __and__(self, other):
        if self.operation is _BitSignalEventGroup.Operation.AND:
            return _BitSignalEventGroup(
                _BitSignalEventGroup.Operation.AND, [*self.events, other]
            )
        else:
            return _BitSignalEventGroup(
                _BitSignalEventGroup.Operation.AND, [self, other]
            )

    def __or__(self, other):
        if self.operation is _BitSignalEventGroup.Operation.OR:
            return _BitSignalEventGroup(
                _BitSignalEventGroup.Operation.OR, [*self.events, other]
            )
        else:
            return _BitSignalEventGroup(
                _BitSignalEventGroup.Operation.OR, [self, other]
            )


@_intrinsic
def rising_edge(_):
    raise AssertionError("rising called outside synthesizable context")


@_intrinsic
def falling_edge(_):
    raise AssertionError("falling called outside synthesizable context")


@_intrinsic
def high_level(_):
    raise AssertionError("high_level called outside synthesizable context")


@_intrinsic
def low_level(_):
    raise AssertionError("low_level called outside synthesizable context")


@_intrinsic_replacement(rising_edge, special_case=False)
def rising_edge_replacement(sig):
    return _BitSignalEvent(sig, _BitSignalEvent.Type.RISING)


@_intrinsic_replacement(falling_edge, special_case=False)
def falling_edge_replacement(sig):
    return _BitSignalEvent(sig, _BitSignalEvent.Type.FALLING)


@_intrinsic_replacement(high_level, special_case=False)
def high_level_replacement(sig):
    return _BitSignalEvent(sig, _BitSignalEvent.Type.HIGH)


@_intrinsic_replacement(low_level, special_case=False)
def low_level_replacement(sig):
    return _BitSignalEvent(sig, _BitSignalEvent.Type.LOW)



from __future__ import annotations

import enum
from typing import Any

from cohdl._core._ir import _repr as ir
from cohdl._core._ir import AccessFlags

from . import _prepare_ast_out as out
from ._traceback import pretty_traceback_active
from cohdl._core._primitive_type import is_primitive


from cohdl._core import (
    _boolean,
    Bit,
    _type_qualifier,
    Signal,
    Temporary,
    Variable,
    Null,
)
from cohdl._core._intrinsic_operations import AssignMode

from cohdl.utility import IdMap, IdSet


class IrGenerator:
    class Mode(enum.Enum):
        CONCURRENT = enum.auto()
        SEQUENTIAL = enum.auto()
        BLOCK = enum.auto()

    @staticmethod
    def convert_sequential(inp: out.Sequential):
        assert (
            not inp.code().returns()
        ), f"return from top level sequential function not possible"

        converter = IrGenerator(IrGenerator.Mode.SEQUENTIAL)
        converter.apply(inp.code(), open_blocks=[converter.code()])

        always_expr = None

        if len(inp.always_expr()) != 0:
            always_converter = IrGenerator(IrGenerator.Mode.CONCURRENT)
            always_converter.apply(
                inp.always_expr(), open_blocks=[always_converter.code()]
            )

            # concurrent block for always assignments
            concurrent = ir.Concurrent("always", always_converter.code(), {}, None)

            temp_replacement = IdMap()

            def find_temporaries(obj, access: AccessFlags):
                if isinstance(obj, Temporary):
                    parent = obj._root

                    if access.is_written():
                        temp_replacement[parent] = Signal[parent.type]()
                    elif parent not in temp_replacement:
                        raise AssertionError(
                            f"always expression cannot inherit temporaries (found {obj._root})"
                        )
                return obj

            def replace_temporaries(obj, access: AccessFlags):
                if isinstance(obj, Temporary):
                    parent = obj._root
                    if parent in temp_replacement:
                        return Signal[obj.type](
                            obj.type(),
                            _root=temp_replacement[parent],
                            _ref_spec=obj._ref_spec,
                        )

                return obj

            # replace temporaries in always expression
            # with signals, this is required, because
            # only signals can be shared between blocks
            # and the result must be available in the
            # sequential context
            concurrent.visit_referenced_objects(find_temporaries)

            concurrent.visit_referenced_objects(replace_temporaries)
            converter.code().visit_referenced_objects(replace_temporaries)

            always_expr = concurrent

        return ir.Sequential(
            inp.name(),
            converter.code(),
            always_expr,
            inp.sensitivity(),
            inp.attributes(),
            inp.source_location(),
        )

    @staticmethod
    def convert_concurrent(inp: out.Concurrent):
        converter = IrGenerator(IrGenerator.Mode.CONCURRENT)
        converter.apply(inp.code(), open_blocks=[converter.code()])

        return ir.Concurrent(
            inp.name(), converter.code(), inp.attributes(), inp.source_location()
        )

    def __init__(
        self,
        mode: IrGenerator.Mode,
    ):
        super().__init__()
        self._mode = mode
        self._code = ir.CodeBlock([], parent=None)
        self._sensitivity = []

    def code(self):
        return self._code

    def add_sensitivity(self, sensitivity):
        assert self._mode is IrGenerator.Mode.SEQUENTIAL
        self._sensitivity.append(sensitivity)

    def _convert_bound(self, inp: out.Statement, open_blocks: list[ir.CodeBlock]):
        for bound in inp.bound_statements():
            open_blocks = self.apply(bound, open_blocks=open_blocks)

        return open_blocks

    _break_result: list[ir.CodeBlock] = []
    _continue_result: list[ir.CodeBlock] = []

    # list of code blocks in the function, that ended in return
    # not open during rest of function
    # must be added to open blocks after end of function
    returned_blocks: list[ir.CodeBlock] = []

    #
    #
    #
    #

    def apply(self, inp: out.Statement, open_blocks: list[ir.CodeBlock]):
        prev_frame = ir.Statement._current_frame

        try:
            ir.Statement._current_frame = inp._frame
        except AttributeError:
            pass

        result = self._apply_impl(inp, open_blocks)

        ir.Statement._current_frame = prev_frame

        return result

    def _apply_impl(self, inp, open_blocks: list[ir.CodeBlock]):
        if isinstance(inp, out.Statement) and not isinstance(
            inp, (out.While, out.Await)
        ):
            open_blocks = self._convert_bound(inp, open_blocks)

        if isinstance(inp, out.Assign):
            target = inp.target()
            value = inp.value()
            mode = inp._mode

            if mode is AssignMode.AUTO:
                if isinstance(target, Signal):
                    mode = AssignMode.NEXT
                elif isinstance(target, Temporary):
                    mode = AssignMode._TEMP
                elif isinstance(target, Variable):
                    mode = AssignMode.VALUE
                else:
                    raise AssertionError("invlaid target")

            if mode is AssignMode.NEXT:
                for block in open_blocks:
                    block.append(ir.SignalAssignment(target, value))
            elif mode is AssignMode.PUSH:
                for block in open_blocks:
                    block.append(ir.SignalPush(target, value))
            elif mode is AssignMode.VALUE:
                for block in open_blocks:
                    block.append(ir.VariableAssignment(target, value))
            elif mode is AssignMode._TEMP:
                if self._mode is IrGenerator.Mode.SEQUENTIAL:
                    for block in open_blocks:
                        block.append(ir.VariableAssignment(target, value))
                else:
                    assert self._mode is IrGenerator.Mode.CONCURRENT
                    for block in open_blocks:
                        block.append(ir.SignalAssignment(target, value))
            else:
                raise AssertionError(f"invalid AssignMode {mode}")

            return open_blocks

        if isinstance(inp, out.SignalAlias):
            for block in open_blocks:
                block.append(ir._SignalAlias(inp._signal, inp._replacement))

            return open_blocks

        if isinstance(inp, out.UnaryOp):
            op = inp._op
            arg = inp._arg
            result = inp.result()

            open_blocks = self.apply(arg, open_blocks=open_blocks)

            for block in open_blocks:
                block.append(ir.UnaryOp(op, arg.result(), result))

            return open_blocks

        if isinstance(inp, out.BinOp):
            op = inp._op
            lhs = inp._lhs
            rhs = inp._rhs
            result = inp.result()

            open_blocks = self.apply(lhs, open_blocks=open_blocks)
            open_blocks = self.apply(rhs, open_blocks=open_blocks)

            for block in open_blocks:
                block.append(ir.BinOp(op, lhs.result(), rhs.result(), result))

            return open_blocks

        if isinstance(inp, out.Compare):
            op = inp._op
            lhs = inp._lhs
            rhs = inp._rhs

            open_blocks = self.apply(lhs, open_blocks=open_blocks)
            open_blocks = self.apply(rhs, open_blocks=open_blocks)

            for block in open_blocks:
                block.append(ir.Compare(op, lhs.result(), rhs.result(), inp.result()))

            return open_blocks

        if isinstance(inp, out.All):
            for block in open_blocks:
                block.append(ir.All(inp._conditions, inp.result()))

            return open_blocks

        if isinstance(inp, out.Any):
            for block in open_blocks:
                block.append(ir.Any(inp._conditions, inp.result()))

            return open_blocks

        if isinstance(inp, out.If):
            test = inp._test
            body = inp._body
            orelse = inp._orelse

            open_blocks = self.apply(test, open_blocks=open_blocks)
            ret_blocks: IdMap[Any, ir.CodeBlock] = IdMap()

            for block in open_blocks:
                code_body = ir.CodeBlock([], parent=block)
                code_orelse = ir.CodeBlock([], parent=block)

                # if statement must be added to the code block before the branches
                # are parsed because there is a special case when await/while statements
                # are the first statements in a sequential and the checks for this case
                # will fail, when the first If statement in a sequential is added after
                # parsing the branches
                block.append(ir.If(test.result(), code_body, code_orelse))

                open_body = self.apply(body, open_blocks=[code_body])
                open_orelse = self.apply(orelse, open_blocks=[code_orelse])

                #
                ##
                #

                def any_transition(root, blocks):
                    if len(blocks) == 0:
                        return True
                    return any(block is not root for block in blocks)

                def all_transition(root, blocks):
                    return all(block is not root for block in blocks)

                any_body = any_transition(code_body, open_body)
                all_body = all_transition(code_body, open_body)
                any_orelse = any_transition(code_orelse, open_orelse)
                all_orelse = all_transition(code_orelse, open_orelse)

                if not any_body and not any_orelse:
                    # no transitions, continue with parent block
                    ret_blocks[block] = block
                elif all_body and all_orelse:
                    # all blocks contain transitions,
                    # continue in new states and not in parent
                    for open in [*open_body, *open_orelse]:
                        ret_blocks[open] = open
                else:
                    if not any_body:
                        # no transition in body, but at least one in orelse
                        # continue in body rather than parent because
                        # otherwise transitions could be overwritten
                        ret_blocks[code_body] = code_body

                        for open in open_orelse:
                            ret_blocks[open] = open
                    elif not any_orelse:
                        # no transitions in orelse, but at least one in body
                        ret_blocks[code_orelse] = code_orelse

                        for open in open_body:
                            ret_blocks[open] = open
                    else:
                        for open in [*open_body, *open_orelse]:
                            ret_blocks[open] = open

            return [*ret_blocks.values()]

        if isinstance(inp, out.Call):
            # store returned blocks of parent Call
            # restore before return
            prev_returned_blocks = IrGenerator.returned_blocks
            own_returned_blocks = []
            IrGenerator.returned_blocks = own_returned_blocks

            # convert function code
            result = self.apply(inp._code, open_blocks=open_blocks)

            # add blocks, that ended with return
            result.extend(own_returned_blocks)

            # restore returned blocks of parent Call
            IrGenerator.returned_blocks = prev_returned_blocks

            return result

            # TODO: check if this works in all cases and improves the generated VHDL

            root_groups: IdMap[ir.CodeBlock, list[ir.CodeBlock]] = IdMap()

            for block in result:
                root = block._root

                if root in root_groups:
                    root_groups[root].append(block)
                else:
                    root_groups[root] = [block]

            if len(root_groups) == 1:
                root_group, *rest = root_groups.values()
                return [ir.CodeBlock.common_block(root_group)]
            return result

        if isinstance(inp, out.Return):
            final = inp._final_bound_statements

            if len(final) != 0:
                # translate possible __exit__ block introduced by with/async with statement
                open_blocks = self.apply(inp._final_bound_statements, open_blocks)

            IrGenerator.returned_blocks.extend(open_blocks)

            # nothing to add to  blocks until end of Call
            # returned_blocks will be added to open_blocks of call
            return []

        if isinstance(inp, out.Await):
            if len(open_blocks) == 0:
                return open_blocks

            ctx = ir.StatemachineContext.get()

            if inp._awaitable_primitive:
                for expr_before in inp._expr_before:
                    open_blocks = self.apply(expr_before, open_blocks=open_blocks)

                if ctx.at_start():
                    # special case for sequential instances with `await` as
                    # first statement. Empty first state is used to avoid
                    # delay of one tick at start of instance.
                    new_state = ctx.first_state()
                else:
                    new_block = ir.CodeBlock([], parent=None)
                    new_state = ir._State(new_block, new_block)
                    ctx.add_state(new_state)

                    for block in open_blocks:
                        # add transition at front of block to allow
                        # contained await expressions to overwrite them
                        block.addfront(ir._Transition(new_state))

                result_list = self.apply(
                    inp.bound_statements(), open_blocks=[new_state.code()]
                )

                # assert first is new_state
                assert len(result_list) == 1

                if isinstance(inp.result(), _boolean._BooleanLiteral):
                    # boolean literals do not require If Statements

                    if inp.result() is _boolean.false:
                        # waiting for cohdl.false stops the statemachine,
                        # the following code is never executed
                        # return empty list, there are no open blocks after await false
                        # because execution stops at that point
                        return []
                    
                    assert inp.result() is _boolean.true, f"internal error: expected boolean literal, got {inp.result()}"
                else: 
                    if_body = ir.CodeBlock([], parent=new_state.open_block())
                    new_state.append(
                        ir.If(
                            inp.result(),
                            if_body,
                            ir.CodeBlock([], parent=new_state.open_block()),
                        )
                    )
                    new_state.set_open_block(if_body)

                return [new_state._open_block]
            else:
                # await expression
                return self.apply(inp.bound_statements(), open_blocks=open_blocks)

        if isinstance(inp, out.While):
            assert (
                inp._test.result() is not False
            ), "internal error: while loops with always false argument should be filtered in the previous compiler stage"

            ctx = ir.StatemachineContext.get()

            if ctx.at_start():
                # special case for sequential instances with `while` as
                # first statement. Empty first state is used to avoid
                # delay of one tick at start of instance.
                new_state = ctx.first_state()
                # add a Nop to mark the state as used
                # so contained statments do not detect
                # the state as empty
                new_state.code().append(ir.Nop())
            else:
                new_block = ir.CodeBlock([], parent=None)
                new_state = ir._State(new_block, new_block)

                for block in open_blocks:
                    # add transition at front of block to allow
                    # contained await expressions to overwrite them
                    block.addfront(ir._Transition(new_state))

                ctx.add_state(new_state)

            # evaluate bound statements before continuing with
            # actual code of while loop
            open_blocks = self._convert_bound(inp, [new_state._open_block])

            assert (
                len(open_blocks) == 1
            ), "internal error: maybe caused by await expression in argument of while loop"
            open_block = open_blocks[0]

            body = ir.CodeBlock([], parent=open_block)
            ret_blocks = []

            prev_continue_result = IrGenerator._continue_result
            prev_break_result = IrGenerator._break_result

            continue_result = []
            break_result = []

            IrGenerator._continue_result = continue_result
            IrGenerator._break_result = break_result

            try:
                # write converted code into body of if statement
                for open_body in self.apply(inp._body, open_blocks=[body]):
                    # add transition to start of while loop
                    # add to front so the transition can be overwritten
                    # by contained await expressions
                    open_body.addfront(ir._Transition(new_state))
            finally:
                IrGenerator._continue_result = prev_continue_result
                IrGenerator._break_result = prev_break_result

            for continue_block in continue_result:
                assert (
                    continue_block.common_block([continue_block, body]) is None
                ), "continue-statement cannot be defined in first state of while-loop (would lead to infinite recursion)"

                if inp._test.result() is True:
                    # for while loops with constant condition
                    # body is the first block in the loop, add statements until
                    # first state change in place of continue
                    continue_block.append(body)
                else:
                    # while loops with runtime variable conditions, continue-statements behave like
                    # break statements if the condition is false. Otherwise the behavior is the same
                    # as a while loop with constant True condition.

                    continue_blocks = self.apply(inp._test, [continue_block])

                    for block in continue_blocks:
                        break_block = ir.CodeBlock([], parent=continue_block)
                        block.append(ir.If(inp._test.result(), body, break_block))
                        ret_blocks.append(break_block)

            # codeblocks, that end with a break statement
            # are the new open blocks to be populated after the while loop
            ret_blocks.extend(break_result)

            if inp._test.result() is True:
                # while loop with constant, true condition
                # exits via break/return are handled earlier
                open_block.append(body)
                return ret_blocks
            else:
                orelse = ir.CodeBlock([], parent=open_block)
                open_block.append(ir.If(inp._test.result(), body, orelse))
                open_block.get_parent_state().set_open_block(orelse)
                return [orelse, *ret_blocks]

        if isinstance(inp, out.Continue):
            # store list of open blocks
            # so while loop can replace continue with statements
            IrGenerator._continue_result.extend(open_blocks)

            # no open blocks after continue since continue is always the last
            # statement in a codeblock
            return []

        if isinstance(inp, out.Break):
            # store list of open blocks
            # so while loop can continue after break
            IrGenerator._break_result.extend(open_blocks)

            # no open blocks after break since break is always the last
            # statement in a codeblock
            return []

        if isinstance(inp, out.Statemachine):
            assert (
                len(open_blocks) == 1
            ), "statemachine instantiated in more than one path"

            parent_block = open_blocks[0]

            ctx = ir.StatemachineContext.enter(inp._name)

            statemachine_end = self.apply(inp._body, open_blocks=[ctx.first_block()])

            parent_block.append(ir.StatemachineContext.finish(statemachine_end))

            return open_blocks

        if isinstance(inp, out.ResetContext):
            for block in open_blocks:
                block.append(ir._ResetContext())
            return open_blocks

        if isinstance(inp, out.ResetPushed):
            for block in open_blocks:
                block.append(ir._ResetPushed())
            return open_blocks

        if isinstance(inp, out.CodeBlock):
            return self.apply(inp.statements(), open_blocks=open_blocks)

        if isinstance(inp, list):
            for stmt in inp:
                open_blocks = self.apply(stmt, open_blocks=open_blocks)

                if stmt.returns_always():
                    break

            return open_blocks

        if isinstance(inp, out.Value):
            return open_blocks

        if isinstance(inp, out.Boolean):
            for block in open_blocks:
                block.append(ir.Boolean(inp.value(), inp.result()))

            return open_blocks

        if isinstance(inp, out.Nop):
            return open_blocks

        if isinstance(inp, out.ResetInstance):
            for block in open_blocks:
                block.append(ir.ResetInstance(inp.get_obj()))

            return open_blocks

        if isinstance(inp, out.IfExpr):
            hook_body = inp._hook_body
            hook_orelse = inp._hook_orelse

            open_blocks = self.apply(inp._body, open_blocks=open_blocks)
            open_blocks = self.apply(inp._orelse, open_blocks=open_blocks)
            open_blocks = self.apply(inp._test, open_blocks=open_blocks)

            for block in open_blocks:
                if hook_body.has_redirect():
                    assert len(hook_body.redirects) == len(hook_orelse.redirects)
                    for branch_body, branch_orelse in zip(
                        hook_body.redirects, hook_orelse.redirects
                    ):
                        test_val = inp._test.result()

                        if issubclass(test_val.type, _boolean._Boolean):
                            true_val = _boolean._Boolean(True)
                        else:
                            assert issubclass(test_val.type, Bit)
                            true_val = Bit(True)

                        block.append(
                            ir.SelectWith(
                                test_val,
                                [
                                    (true_val, branch_body.source),
                                ],
                                default=branch_orelse.source,
                                result=branch_body.target,
                            )
                        )

            return open_blocks

        if isinstance(inp, out.SelectWith):
            branch_hooks = inp._branch_hooks
            first_hook = branch_hooks[0]

            for block in open_blocks:
                if first_hook.has_redirect():
                    for nr, redirect in enumerate(first_hook.redirects):
                        branches = [
                            (cond, hook.redirects[nr].source)
                            for cond, hook in zip(inp._conditions, branch_hooks)
                        ]

                        if inp._default_hook is not None:
                            block.append(
                                ir.SelectWith(
                                    inp._arg,
                                    branches,
                                    inp._default_hook.redirects[nr].source,
                                    redirect.target,
                                )
                            )
                        else:
                            block.append(
                                ir.SelectWith(
                                    inp._arg,
                                    branches,
                                    None,
                                    redirect.target,
                                )
                            )

            return open_blocks

        if isinstance(inp, out.CondSelect):

            def gen_case_when(
                value: out.Expression,
                conds: list[out.Value],
                bodies: list[out.CodeBlock],
                default: out.CodeBlock | None,
                open_blocks: list[ir.CodeBlock],
            ):
                # convert parameter bodies to internal representation
                def gen_bodies(parent_block):
                    ret = []

                    for code in bodies:
                        code_block = ir.CodeBlock([], parent=parent_block)
                        converted = self.apply(code, open_blocks=[code_block])

                        if len(converted) != 1 or converted[0] is not code_block:
                            # generating case-when not possible, when a branch contains a transition
                            # return None to fall back to if-else implementation
                            return None

                        ret.append(code_block)
                    return ret

                conditions = [expr.result() for expr in conds]

                for block in open_blocks:
                    new_blocks = self.apply(value, open_blocks=[block])

                    if default is None:
                        for new_block in new_blocks:
                            ir_bodies = gen_bodies(new_block)

                            if ir_bodies is None:
                                return None

                            new_block.append(
                                ir.CaseWhen(
                                    value.result(),
                                    [
                                        (cond, body)
                                        for cond, body in zip(conditions, ir_bodies)
                                    ],
                                    None,
                                )
                            )
                    else:
                        for new_block in new_blocks:
                            default_body = ir.CodeBlock([], parent=block)
                            converted = self.apply(default, open_blocks=[default_body])

                            if len(converted) != 1 or converted[0] is not default_body:
                                return None

                            ir_bodies = gen_bodies(new_block)

                            if ir_bodies is None:
                                return None

                            new_block.append(
                                ir.CaseWhen(
                                    value.result(),
                                    [
                                        (cond.result(), body)
                                        for cond, body in zip(conds, ir_bodies)
                                    ],
                                    default_body,
                                )
                            )

                return open_blocks

            def try_gen_case_when(
                inp: out.CondSelect,
                open_blocks: list[ir.CodeBlock],
            ):
                if inp.returns():
                    # fallback to if-else implementation
                    return None

                lhs_map = IdMap()
                rhs_map = IdMap()

                lhs_expr = []
                rhs_expr = []

                bodies = []

                for expr, body in inp._branches:
                    if body.contains_break() or body.contains_continue():
                        # handling of state transitions and break/continue statements is complex
                        # better opt out of case-when statement and instead use already implemented
                        # if-else chain
                        return None

                    if isinstance(expr, out.Compare) and expr._op is expr.Operator.EQ:

                        lhs = expr._lhs
                        rhs = expr._rhs

                        lhs_expr.append(lhs)
                        rhs_expr.append(rhs)

                        lhs_map.map_self(lhs.result())
                        rhs_map.map_self(rhs.result())

                        bodies.append(body)
                    else:
                        return None

                def check_value(map: IdMap):
                    # same value is used in every equality check
                    return len(map) == 1

                def check_branches(map: IdMap):
                    for cond in map.values():
                        assert is_primitive(_type_qualifier.TypeQualifier.decay(cond))

                        if isinstance(cond, _type_qualifier.TypeQualifier):
                            return False
                    return True

                if check_value(lhs_map) and check_branches(rhs_map):
                    return gen_case_when(
                        lhs_expr[0], rhs_expr, bodies, inp._default, open_blocks
                    )

                if check_value(rhs_map) and check_branches(lhs_map):
                    return gen_case_when(
                        rhs_expr[0], lhs_expr, bodies, inp._default, open_blocks
                    )

                return None

            if len(inp._branches) == 0:
                if inp._default is None:
                    return open_blocks
                return self.apply(inp._default, open_blocks=open_blocks)

            #
            # construct a case when statement if possible
            #

            case_when = try_gen_case_when(inp, open_blocks)

            if case_when is not None:
                return case_when

            #
            # fallback to nested if statements
            #

            if inp.returns():
                assert inp.returns_always()

            if inp._default is not None:
                rewritten = inp._default
            else:
                rewritten = out.CodeBlock([])

            for branch in inp._branches[::-1]:
                expr, code = branch

                rewritten = out.If(expr, code, out.CodeBlock([rewritten]))

            return self.apply(rewritten, open_blocks=open_blocks)

        if isinstance(inp, out.Assert):
            for block in open_blocks:
                result_blocks = self.apply(inp._cond, open_blocks=[block])

                assert (
                    len(result_blocks) == 1 and result_blocks[0] is block
                ), "transistions are not allowed in assertation test"

                msg = inp._msg

                if msg is not None:
                    assert isinstance(msg, out.Expression)
                    msg = msg.result()

                block.append(ir.Assert(inp._cond.result(), msg))

            return open_blocks

        if isinstance(inp, out.InlineCode):
            result = inp.result()

            if isinstance(result, out.InlineCode):
                result = None

            for block in open_blocks:
                code = ir.InlineCode(inp.options, result)
                block.append(code)

            return open_blocks

        if isinstance(inp, out.Comment):
            for block in open_blocks:
                block.append(ir.Comment(inp.lines))

            return open_blocks

        raise AssertionError(f"cannot convert {inp}")


#
#
#
# instance converters
#
#
#


class ConvertInstance:
    def __init__(
        self, template: IdMap[out.EntityTemplate, ir.EntityTemplate] | None = None
    ) -> None:
        super().__init__()

        self._entity_templates: IdMap[out.EntityTemplate, ir.EntityTemplate] = (
            IdMap() if template is None else template
        )

    def lookup_template(self, source: out.EntityTemplate) -> ir.EntityTemplate | None:
        if source in self._entity_templates:
            return self._entity_templates[source]
        return None

    def add_template(self, source: out.EntityTemplate, result: ir.EntityTemplate):
        self._entity_templates[source] = result

    @staticmethod
    def detect_uninitialized_temporaries(ctx: ir.Context):
        invalid_temporaries = set()
        written_temporaries = set()

        def check_used_temporaries(obj, access: AccessFlags):
            if access is AccessFlags.READ:
                if isinstance(obj, Temporary):
                    # This assertion ensures, that all temporaries are initialized
                    # before they are used. Check your code for temporaries,
                    # that are defined in a if or match statement and used outside of it.
                    #
                    # In the following example the temporary 'var' is invalid at the location
                    # of the assignment to 'some_output' because it is only initialized when
                    # 'some_signal' is true.
                    #
                    # >>> if some_signal:
                    # >>>     var = inp_a | inp_b
                    # >>> some_output <<= var        # error occurs in this line
                    root_id = id(obj._root)
                    assert (
                        root_id not in invalid_temporaries
                    ), "temporary might not be initialized"

                    assert (
                        root_id in written_temporaries
                    ), "temporary read before it was written (defined outside current context?)"
            elif access is AccessFlags.WRITE:
                written_temporaries.add(id(obj._root))

            return obj

        def search_invalid_temporaries(code: ir.CodeBlock):
            nonlocal invalid_temporaries

            local_temporaries = set()

            for stmt in code._content:
                if isinstance(stmt, ir.If):
                    check_used_temporaries(stmt._test, AccessFlags.READ)

                    # Find all temporaries defined in body and mark them
                    # as invalid because they might not be defined
                    body_temporaries = search_invalid_temporaries(stmt._body)
                    invalid_temporaries |= body_temporaries

                    # Find all temporaries defined in else branch and mark
                    # them as invalid because they might not be defined.
                    else_temporaries = search_invalid_temporaries(stmt._orelse)
                    invalid_temporaries |= else_temporaries

                    always_defined = body_temporaries & else_temporaries

                    # When a temporary is defined in both blocks
                    # if is always defined and thus valid.
                    invalid_temporaries.difference_update(always_defined)
                    local_temporaries |= always_defined

                elif isinstance(stmt, ir.CodeBlock):
                    local_temporaries |= search_invalid_temporaries(stmt)
                elif isinstance(stmt, ir.CaseWhen):
                    check_used_temporaries(stmt._value, AccessFlags.READ)

                    always_defined = None

                    for branch_cond, branch_code in stmt._branches:
                        check_used_temporaries(branch_cond, AccessFlags.READ)
                        branch_temporaries = search_invalid_temporaries(branch_code)
                        invalid_temporaries |= branch_temporaries

                        if always_defined is None:
                            always_defined = branch_temporaries
                        else:
                            always_defined.difference_update(branch_temporaries)

                    if stmt._default is not None:
                        default_temporaries = search_invalid_temporaries(stmt._default)
                        invalid_temporaries |= default_temporaries

                        if always_defined is None:
                            always_defined = branch_temporaries
                        else:
                            always_defined.difference_update(branch_temporaries)

                    invalid_temporaries.difference_update(always_defined)
                    local_temporaries |= always_defined

                else:
                    ir._visit_referenced_objects(stmt, check_used_temporaries)

                    if isinstance(stmt, ir.Expression) and isinstance(
                        stmt._result, Temporary
                    ):
                        root = stmt._result._root

                        if not root._maybe_uninitialized:
                            root_id = id(root)

                            local_temporaries.add(root_id)
                            invalid_temporaries.discard(root_id)

                    elif isinstance(stmt, ir.VariableAssignment) and isinstance(
                        stmt._target, Temporary
                    ):
                        root = stmt._target._root

                        if not root._maybe_uninitialized:
                            root_id = id(root)
                            local_temporaries.add(root_id)
                            invalid_temporaries.discard(root_id)

            return local_temporaries

        search_invalid_temporaries(ctx.code())

    @staticmethod
    def cleanup_unused(ctx: ir.Context):
        used_temporaries = IdSet()

        def find_used_temp(obj, access: AccessFlags):
            if access.is_read() and isinstance(obj, Temporary):
                root = obj._root
                used_temporaries.add(root)

            return obj

        ctx.visit_referenced_objects(find_used_temp)

        def remove_unused_assignments(stmt):
            if isinstance(stmt, ir.Expression) and isinstance(stmt.result(), Temporary):
                root = stmt.result()._root

                if root not in used_temporaries:
                    return ir.CodeBlock([], None)
            elif isinstance(stmt, ir.VariableAssignment) and isinstance(
                stmt._target, Temporary
            ):
                root = stmt._target._root

                if root not in used_temporaries:
                    return ir.CodeBlock([], None)

            return stmt

        ctx.visit(remove_unused_assignments)

        return ctx

    @staticmethod
    def cleanup_bool_cast(ctx: ir.Context):
        # The previous stages of the compiler produce some
        # redundant casts from Temporary[bool] to Temporary[bool].
        # This function removes these casts and replaces all uses
        # of the cast result with the cast input.
        # This is a purely cosmetic operation to make the generated
        # HDL more readable.

        replacement_map = IdMap()

        def search_unneeded_bool_casts(stmt):
            if isinstance(stmt, ir.Boolean):
                source = stmt._arg
                target = stmt._result
                if (
                    isinstance(source, Temporary)
                    and isinstance(target, Temporary)
                    and source._root is source
                    and target._root is target
                    and source.type is _boolean.boolean
                    and target.type is _boolean.boolean
                ):
                    replacement_map[target] = source
                    return ir.Nop()

            return stmt

        ctx.visit(search_unneeded_bool_casts)

        def replace_temporaries(obj, access):
            if obj in replacement_map:
                return replacement_map[obj]

            return obj

        ctx.visit_referenced_objects(replace_temporaries)
        return ctx

    #
    #
    #

    def apply(self, inp):
        try:
            if isinstance(inp, out.EntityTemplate):
                ir_template = self.lookup_template(inp)

                if ir_template is None:
                    ir_template = ir.EntityTemplate(
                        inp._info,
                        [self.apply(block) for block in inp.subblocks()],
                        [self.apply(ctx) for ctx in inp.contexts()],
                    )

                    self.add_template(inp, ir_template)

                return ir_template

            if isinstance(inp, out.Entity):
                template = self.apply(inp.template())

                return ir.Entity(
                    template,
                    inp._info.name,
                    inp.port_definitions(),
                    inp.generic_definitions(),
                )
            if isinstance(inp, out.Block):
                return ir.Block(
                    "<BLOCK>",
                    [self.apply(subinst) for subinst in inp.subblocks()],
                    [self.apply(ctx) for ctx in inp.contexts()],
                    inp._attributes,
                )

            if isinstance(inp, out.Concurrent):
                result = IrGenerator.convert_concurrent(inp)

                written_temporaries = set()

                def check_variables_and_temporaries(obj, access):
                    assert not isinstance(
                        obj, Variable
                    ), "variables cannot be used in concurrent contexts"

                    if isinstance(obj, Temporary):
                        if access is AccessFlags.READ:
                            assert (
                                id(obj._root) in written_temporaries
                            ), "temporary read before it was written (defined outside current context?)"
                        else:
                            written_temporaries.add(id(obj._root))

                    return obj

                result.visit_referenced_objects(check_variables_and_temporaries)

                if result.attributes.get("cleanup_unused", True):
                    result = ConvertInstance.cleanup_unused(result)

                if result.attributes.get("zero_init_temporaries", False):
                    # only used for unit tests
                    # (required because ghdl terminates with an overflow error
                    # when uninitialized integer values are used)

                    def visit_obj(obj, access):
                        if isinstance(obj, Temporary):
                            obj._default = obj.type(Null)
                        return obj

                    result.visit_referenced_objects(visit_obj)
                return result

            if isinstance(inp, out.Sequential):
                result = IrGenerator.convert_sequential(inp)

                ConvertInstance.detect_uninitialized_temporaries(result)

                if result.attributes.get("cleanup_unused", True):
                    result = ConvertInstance.cleanup_unused(result)
                if result.attributes.get("cleanup_bool_cast", True):
                    result = ConvertInstance.cleanup_bool_cast(result)
                if result.attributes.get("zero_init_temporaries", False):
                    # only used for unit tests
                    # (required because ghdl terminates with an overflow error
                    # when uninitialized integer values are used)

                    def visit_obj(obj, access):
                        if isinstance(obj, Temporary):
                            obj._default = obj.type(Null)
                        return obj

                    result.visit_referenced_objects(visit_obj)
                return result

            raise AssertionError(f"cannot convert {inp}")
        except ir.VisitException as err:
            if pretty_traceback_active():
                src = err.src_statement

                if src._frame is not None:
                    src._frame.apply_to_exception(err.original)

                raise err.original
            raise



from __future__ import annotations

from cohdl._core._intrinsic import _intrinsic


class InlineCode:
    expr_type = None

    class Node:
        ...

    class Text(Node):
        def __init__(self, text: str):
            self.text = text

    class Object(Node):
        def __init__(self, obj, read: bool):
            self.obj = obj
            self.read = read

    class SubCode(Node):
        def __init__(self, options: list[InlineCode]) -> None:
            self.options = options

    def __init__(self, content: list[InlineCode.Node]):
        self.content = content

    def copy(self):
        return type(self)(self.obj, self.read)

    def post_process(self, text):
        return text

    @_intrinsic
    def __class_getitem__(cls, _expr_type):
        class _Inline(cls):
            expr_type = _expr_type

        return _Inline


class InlineRaw(InlineCode):
    ...


class InlineVhdl(InlineCode):
    ...



import unittest
import random

from cohdl import Unsigned, BitOrder


class UnsignedTester(unittest.TestCase):
    def test_init(self):
        for w in (1, 2, 3, 15, 37, 128):
            for x in [random.randint(0, 2**w - 1) for _ in range(25)]:
                val = Unsigned[w](x)

                for cmp in (val, x, Unsigned[w](val)):
                    self.assertTrue(val == cmp)
                    self.assertTrue(cmp == val)
                    self.assertTrue(cmp - val == 0)
                    self.assertTrue(val - cmp == 0)
                    self.assertTrue(0 == cmp - val)
                    self.assertTrue(0 == val - cmp)

                    self.assertFalse(cmp != val)
                    self.assertFalse(val != cmp)

                    self.assertFalse(cmp < val)
                    self.assertFalse(val < cmp)
                    self.assertFalse(cmp > val)
                    self.assertFalse(val > cmp)

                    self.assertTrue(cmp <= val)
                    self.assertTrue(val <= cmp)
                    self.assertTrue(cmp >= val)
                    self.assertTrue(val >= cmp)

    def test_init_asserts(self):
        for w in (1, 2, 3, 15, 1024):
            # TODO: test or remove order
            for order in (BitOrder.DOWNTO, BitOrder.UPTO):
                self.assertRaises(AssertionError, Unsigned[w], -1)
                self.assertRaises(AssertionError, Unsigned[w], -17)
                self.assertRaises(AssertionError, Unsigned[w], 2**w)
                self.assertRaises(
                    AssertionError,
                    Unsigned[w],
                    2**w + 53225265256,
                )
                self.assertRaises(
                    AssertionError,
                    Unsigned[w],
                    "0" * (w - 1),
                )
                self.assertRaises(
                    AssertionError,
                    Unsigned[w],
                    "00000" * (w + 1),
                )



import unittest

from cohdl import Bit, BitState


class BitStateTester(unittest.TestCase):
    def test_bool_conv(self):
        self.assertTrue(BitState.HIGH)
        self.assertFalse(BitState.LOW)

    def test_inv(self):
        self.assertTrue(~BitState.LOW is BitState.HIGH)
        self.assertTrue(~BitState.HIGH is BitState.LOW)

    def test_binop(self):

        h = BitState.HIGH
        l = BitState.LOW
        # reveal_type(h | l)

        # binary or
        self.assertTrue(l | l is l)
        self.assertTrue(l | h is h)
        self.assertTrue(h | l is h)
        self.assertTrue(h | h is h)

        # binary and
        self.assertTrue(l & l is l)
        self.assertTrue(l & h is l)
        self.assertTrue(h & l is l)
        self.assertTrue(h & h is h)

        # binary xor
        self.assertTrue(l ^ l is l)
        self.assertTrue(l ^ h is h)
        self.assertTrue(h ^ l is h)
        self.assertTrue(h ^ h is l)

    def test_tostr(self):
        self.assertTrue(str(BitState.LOW) == "0")
        self.assertTrue(str(BitState.HIGH) == "1")
        self.assertTrue(str(BitState.UNINITIALZED) == "U")
        self.assertTrue(str(BitState.UNKNOWN) == "X")
        self.assertTrue(str(BitState.HIGH_IMPEDANCE) == "Z")

    def test_fromstr(self):
        self.assertTrue(BitState.from_str("0") is BitState.LOW)
        self.assertTrue(BitState.from_str("1") is BitState.HIGH)
        self.assertTrue(BitState.from_str("U") is BitState.UNINITIALZED)
        self.assertTrue(BitState.from_str("X") is BitState.UNKNOWN)
        self.assertTrue(BitState.from_str("-") is BitState.DONT_CARE)

        self.assertRaises(AssertionError, BitState.construct, "")
        self.assertRaises(AssertionError, BitState.construct, "00")
        self.assertRaises(AssertionError, BitState.construct, "asdf")
        self.assertRaises(AssertionError, BitState.construct, "x")
        self.assertRaises(AssertionError, BitState.construct, "u")
        self.assertRaises(AssertionError, BitState.construct, "_")

    def test_construct(self):
        self.assertTrue(BitState.construct(False) is BitState.LOW)
        self.assertTrue(BitState.construct(True) is BitState.HIGH)

        self.assertTrue(BitState.construct(0) is BitState.LOW)
        self.assertTrue(BitState.construct(1) is BitState.HIGH)
        self.assertRaises(AssertionError, BitState.construct, 2)
        self.assertRaises(AssertionError, BitState.construct, -1)
        self.assertRaises(AssertionError, BitState.construct, -100)
        self.assertRaises(AssertionError, BitState.construct, 1000)

        self.assertTrue(BitState.construct("0") is BitState.LOW)
        self.assertTrue(BitState.construct("1") is BitState.HIGH)
        self.assertTrue(BitState.construct("U") is BitState.UNINITIALZED)
        self.assertTrue(BitState.construct("X") is BitState.UNKNOWN)
        self.assertTrue(BitState.construct("-") is BitState.DONT_CARE)

        self.assertRaises(AssertionError, BitState.construct, "")
        self.assertRaises(AssertionError, BitState.construct, "00")
        self.assertRaises(AssertionError, BitState.construct, "asdf")
        self.assertRaises(AssertionError, BitState.construct, "x")
        self.assertRaises(AssertionError, BitState.construct, "u")
        self.assertRaises(AssertionError, BitState.construct, "_")


class BitTester(unittest.TestCase):
    def test_init(self):
        def with_arg(arg, expected_bool: bool, expected_state: BitState):
            bit = Bit(arg)

            self.assertTrue(bool(bit) is expected_bool)
            self.assertTrue(bit.get() is expected_state)
            self.assertFalse(bit is Bit(arg))

            cpy = bit.copy()
            self.assertFalse(bit is cpy)
            self.assertTrue(bit.get() is cpy.get())

        with_arg(False, False, BitState.LOW)
        with_arg(True, True, BitState.HIGH)

        with_arg(0, False, BitState.LOW)
        with_arg(1, True, BitState.HIGH)

        with_arg("0", False, BitState.LOW)
        with_arg("1", True, BitState.HIGH)

        with_arg(BitState.LOW, False, BitState.LOW)
        with_arg(BitState.HIGH, True, BitState.HIGH)

        with_arg(Bit(BitState.LOW), False, BitState.LOW)
        with_arg(Bit(BitState.HIGH), True, BitState.HIGH)

    def test_copy(self):
        def copy_bit(bit):
            cpy = bit.copy()

            self.assertTrue(cpy is not bit)
            self.assertTrue(cpy.get() is bit.get())
            self.assertTrue(cpy == bit)

        args = [False, True, 0, 1, "0", "1"]

        for default_arg in args:
            bit = Bit(default_arg)
            copy_bit(bit)

    def test_invert(self):
        bit = Bit(False)

        a = bit.invert()
        self.assertTrue(isinstance(a, Bit))
        self.assertTrue(bit.get() is BitState.LOW)
        self.assertTrue(a.get() is BitState.HIGH)

        b = ~bit
        self.assertTrue(isinstance(b, Bit))
        self.assertTrue(bit.get() is BitState.LOW)
        self.assertTrue(b.get() is BitState.HIGH)

        a = a.invert()
        b = ~b

        self.assertTrue(a.get() is BitState.LOW)
        self.assertTrue(b.get() is BitState.LOW)

    def test_binop(self):
        def helper(result: Bit, expected):
            self.assertTrue(isinstance(result, Bit))
            self.assertTrue(bool(result) == expected)

        l = Bit(False)
        h = Bit(True)

        helper(l | l, False)
        helper(l | h, True)
        helper(h | l, True)
        helper(h | h, True)

        helper(l & l, False)
        helper(l & h, False)
        helper(h & l, False)
        helper(h & h, True)

        helper(l ^ l, False)
        helper(l ^ h, True)
        helper(h ^ l, True)
        helper(h ^ h, False)

    def test_cmp(self):
        l = Bit(False)

        self.assertTrue(Bit(False) == Bit(False))
        self.assertTrue(Bit(True) == Bit(True))
        self.assertFalse(Bit(False) == Bit(True))
        self.assertFalse(Bit(True) == Bit(False))

        self.assertFalse(Bit(False) != Bit(False))
        self.assertFalse(Bit(True) != Bit(True))
        self.assertTrue(Bit(False) != Bit(True))
        self.assertTrue(Bit(True) != Bit(False))

        self.assertTrue(Bit(False) == BitState.LOW)
        self.assertTrue(Bit(True) == BitState.HIGH)
        self.assertFalse(Bit(False) == BitState.HIGH)
        self.assertFalse(Bit(True) == BitState.LOW)

        self.assertFalse(Bit(False) != BitState.LOW)
        self.assertFalse(Bit(True) != BitState.HIGH)
        self.assertTrue(Bit(False) != BitState.HIGH)
        self.assertTrue(Bit(True) != BitState.LOW)

        self.assertTrue(BitState.LOW == Bit(False))
        self.assertTrue(BitState.HIGH == Bit(True))
        self.assertFalse(BitState.LOW == Bit(True))
        self.assertFalse(BitState.HIGH == Bit(False))

        self.assertFalse(BitState.LOW != Bit(False))
        self.assertFalse(BitState.HIGH != Bit(True))
        self.assertTrue(BitState.LOW != Bit(True))
        self.assertTrue(BitState.HIGH != Bit(False))



from __future__ import annotations

from cohdl.std._context import SequentialContext
from cohdl.std.axi._axi4_channel import Channel
from cohdl.std._core_utility import Mask, stretch

import cohdl


class Axi4Light:
    class RespConstants:
        OKAY = "00"
        EXOKAY = "01"
        SLVERR = "10"
        DECERR = "11"

    class WrAddr(Channel):
        def __init__(self, valid, ready, awaddr, awprot, **kwargs):
            super().__init__(valid, ready, awaddr=awaddr, awprot=awprot, **kwargs)

            self.awaddr = awaddr
            self.awprot = awprot

    class WrData(Channel):
        def __init__(self, valid, ready, wdata, wstrb, **kwargs):
            super().__init__(valid, ready, wdata=wdata, wstrb=wstrb, **kwargs)

            self.wdata = wdata
            self.wstrb = wstrb

    class WrResp(Channel):
        def __init__(self, valid, ready, bresp, **kwargs):
            super().__init__(valid, ready, bresp=bresp, **kwargs)

            self.bresp = bresp

    class RdAddr(Channel):
        def __init__(self, valid, ready, araddr, arprot, **kwargs):
            super().__init__(valid, ready, araddr=araddr, arprot=arprot, **kwargs)

            self.araddr = araddr
            self.arprot = arprot

    class RdData(Channel):
        def __init__(self, valid, ready, rdata, rresp, **kwargs):
            super().__init__(valid, ready, rdata=rdata, rresp=rresp, **kwargs)

            self.rdata = rdata
            self.rresp = rresp

    @staticmethod
    def signal(
        clk,
        reset,
        addr_width,
        data_width=32,
        strb_width=4,
        prot_width=3,
        resp_width=2,
        prefix="",
    ):
        S = cohdl.Signal
        Bit = cohdl.Bit
        BitVector = cohdl.BitVector
        Null = cohdl.Null

        awaddr = Axi4Light.WrAddr(
            S[Bit](Null, name=f"{prefix}awvalid"),
            S[Bit](Null, name=f"{prefix}awready"),
            S[BitVector[addr_width]](Null, name=f"{prefix}awaddr"),
            (
                S[BitVector[prot_width]](Null, name=f"{prefix}awprot")
                if prot_width is not None
                else None
            ),
        )

        wrdata = Axi4Light.WrData(
            S[Bit](Null, name=f"{prefix}wvalid"),
            S[Bit](Null, name=f"{prefix}wready"),
            S[BitVector[data_width]](Null, name=f"{prefix}wdata"),
            S[BitVector[strb_width]](Null, name=f"{prefix}wstrb"),
        )

        bresp = Axi4Light.WrResp(
            S[Bit](Null, name=f"{prefix}bvalid"),
            S[Bit](Null, name=f"{prefix}bready"),
            S[BitVector[resp_width]](Null, name=f"{prefix}bresp"),
        )

        rdaddr = Axi4Light.RdAddr(
            S[Bit](Null, name=f"{prefix}arvalid"),
            S[Bit](Null, name=f"{prefix}arready"),
            S[BitVector[addr_width]](Null, name=f"{prefix}araddr"),
            (
                S[BitVector[prot_width]](Null, name=f"{prefix}arprot")
                if prot_width is not None
                else None
            ),
        )

        rddata = Axi4Light.RdData(
            S[Bit](Null, name=f"{prefix}rvalid"),
            S[Bit](Null, name=f"{prefix}rready"),
            S[BitVector[data_width]](Null, name=f"{prefix}rdata"),
            S[BitVector[resp_width]](Null, name=f"{prefix}rresp"),
        )

        return Axi4Light(clk, reset, awaddr, wrdata, bresp, rdaddr, rddata)

    def __init__(
        self,
        clk,
        reset,
        wraddr: Axi4Light.WrAddr,
        wrdata: Axi4Light.WrData,
        wrresp: Axi4Light.WrResp,
        rdaddr: Axi4Light.RdAddr,
        rddata: Axi4Light.RdData,
    ):
        self.clk = clk
        self.reset = reset
        self.wraddr = wraddr
        self.wrdata = wrdata
        self.wrresp = wrresp
        self.rdaddr = rdaddr
        self.rddata = rddata

        self._addr_width = rdaddr.araddr.width
        self._data_width = rddata.rdata.width

    def addr_width(self):
        return self._addr_width

    def data_width(self):
        return self._data_width

    async def read_word(self, addr, prot=cohdl.Null):
        if self.rdaddr.arprot is None:
            assert (
                prot is cohdl.Null
            ), "arprot not provided, interface does not support `prot` argument"
        else:
            self.rdaddr.arprot <<= prot

        self.rdaddr.araddr.unsigned <<= addr

        self.rdaddr.valid <<= True

        await self.rdaddr.ready
        self.rdaddr.valid <<= False

        self.rddata.ready <<= True
        await self.rddata.valid
        self.rddata.ready <<= False

        return self.rddata.rdata, self.rddata.rresp

    async def write_word(self, addr, data, strb=cohdl.Full, prot=cohdl.Null):
        if self.wraddr.awprot is None:
            assert (
                prot is cohdl.Null
            ), "awprot not provided, interface does not support `prot` argument"
        else:
            self.wraddr.awprot <<= prot

        self.wraddr.awaddr.unsigned <<= addr
        self.wraddr.valid <<= True

        self.wrdata.wdata <<= data
        self.wrdata.wstrb <<= strb
        self.wrdata.valid <<= True

        ack_addr = cohdl.Variable(self.wraddr.ready)
        ack_data = cohdl.Variable(self.wrdata.ready)

        while True:
            ack_addr @= self.wraddr.ready | ack_addr
            ack_data @= self.wrdata.ready | ack_data

            self.wraddr.valid <<= ~ack_addr
            self.wrdata.valid <<= ~ack_data

            if ack_addr & ack_data:
                break

        self.wrresp.ready <<= True
        await self.wrresp.valid
        self.wrresp.ready <<= False

        return self.wrresp.bresp

    async def handle_reads(self, read_handler, is_async=False):
        while True:
            self.rdaddr.ready <<= True
            await self.rdaddr.valid
            self.rdaddr.ready <<= False

            if is_async:
                data, resp = await read_handler(self.rdaddr.araddr, self.rdaddr.arprot)
            else:
                data, resp = read_handler(self.rdaddr.araddr, self.rdaddr.arprot)

            self.rddata.rdata <<= data
            self.rddata.rresp <<= resp
            self.rddata.valid <<= True

            await self.rddata.ready
            self.rddata.valid <<= False

            continue

    async def handle_writes(self, write_handler, is_async=False):
        while True:
            self.wraddr.ready <<= False
            self.wrdata.ready <<= False
            self.wrresp.valid <<= False

            if self.wraddr.valid & self.wrdata.valid:
                if is_async:
                    resp = await write_handler(
                        self.wraddr.awaddr,
                        self.wraddr.awprot,
                        self.wrdata.wdata,
                        self.wrdata.wstrb,
                    )
                else:
                    resp = write_handler(
                        self.wraddr.awaddr,
                        self.wraddr.awprot,
                        self.wrdata.wdata,
                        self.wrdata.wstrb,
                    )

                self.wraddr.ready <<= True
                self.wrdata.ready <<= True

                self.wrresp.bresp <<= resp
                self.wrresp.valid <<= True

                while ~self.wrresp.ready:
                    self.wraddr.ready <<= False
                    self.wrdata.ready <<= False

                self.wraddr.ready <<= False
                self.wrdata.ready <<= False
                self.wrresp.valid <<= False

    class _ReadRequest:
        def __init__(self, addr, prot):
            self.addr = addr
            self.prot = prot

    async def await_read_request(self):
        self.rdaddr.ready <<= True
        await self.rdaddr.valid
        self.rdaddr.ready <<= False
        return Axi4Light._ReadRequest(self.rdaddr.araddr, self.rdaddr.arprot)

    async def send_read_resp(self, data, resp=cohdl.Null):
        self.rddata.rdata <<= data
        self.rddata.rresp <<= resp
        self.rddata.valid <<= True
        await self.rddata.ready
        self.rddata.valid <<= False

    class _WriteRequest:
        def __init__(self, addr, prot, data, strb):
            self.addr = addr
            self.prot = prot
            self.data = data
            self.strb = strb

    async def await_write_request(self):
        awready = self.wraddr.ready
        awvalid = self.wraddr.valid

        wready = self.wrdata.ready
        wvalid = self.wrdata.valid

        addr_latched = cohdl.Variable[cohdl.Bit](False)
        data_latched = cohdl.Variable[cohdl.Bit](False)

        awready <<= True
        wready <<= True

        while True:
            if ~addr_latched:
                addr_buffer = cohdl.Signal(self.wraddr.awaddr, maybe_uninitialized=True)
                prot_buffer = cohdl.Signal(self.wraddr.awprot, maybe_uninitialized=True)
            if ~data_latched:
                data_buffer = cohdl.Signal(self.wrdata.wdata, maybe_uninitialized=True)
                strb_buffer = cohdl.Signal(self.wrdata.wstrb, maybe_uninitialized=True)

            addr_latched @= addr_latched | awvalid
            data_latched @= data_latched | wvalid

            awready <<= ~addr_latched
            wready <<= ~data_latched

            if addr_latched & data_latched:
                break

        return Axi4Light._WriteRequest(
            addr_buffer, prot_buffer, data_buffer, strb_buffer
        )

    async def send_write_response(self, resp=cohdl.Null):
        self.wrresp.bresp <<= resp
        self.wrresp.valid <<= True
        await self.wrresp.ready
        self.wrresp.valid <<= False

    @staticmethod
    def mask_wstrb(old_data, new_data, wstrb):
        byte_0 = new_data[7:0] if wstrb[0] else old_data[7:0]
        byte_1 = new_data[15:8] if wstrb[1] else old_data[15:8]
        byte_2 = new_data[23:16] if wstrb[2] else old_data[23:16]
        byte_3 = new_data[31:24] if wstrb[3] else old_data[31:24]

        return byte_3 @ byte_2 @ byte_1 @ byte_0

    def connect_addr_map(self, addr_map):
        from cohdl.std.reg import RegisterTools
        from cohdl.std._core_utility import as_awaitable

        assert isinstance(addr_map, RegisterTools.AddrMap), "addr_map is not a AddrMap"
        assert addr_map._word_width_() == 32, "word width of addr_map object must be 32"
        assert (
            addr_map._register_tools_._addr_unit_width_ == 8
        ), "unit width of addr_map object must be 8"

        ctx = SequentialContext(self.clk, self.reset)
        addr_map._implement_synthesizable_contexts_(ctx)

        regs = addr_map._flatten_()

        readable_regs = [reg for reg in regs if reg._readable_]
        writable_regs = [reg for reg in regs if reg._writable_]

        @ctx
        async def proc_read():
            while True:
                request = await self.await_read_request()
                assert isinstance(request.addr, cohdl.Signal)

                result = cohdl.Variable[cohdl.BitVector[32]](cohdl.Null)

                for reg in readable_regs:
                    if reg._contains_addr_(request.addr.unsigned):
                        result @= await as_awaitable(
                            reg._basic_read_, request.addr.unsigned, None
                        )
                        break

                await self.send_read_resp(result)
                continue

        @ctx
        async def proc_write():
            while True:
                request = await self.await_write_request()
                assert isinstance(request.addr, cohdl.Signal)
                mask = cohdl.Variable(stretch(request.strb, 8))

                for reg in writable_regs:
                    if reg._contains_addr_(request.addr.unsigned):
                        await as_awaitable(
                            reg._basic_write_,
                            request.addr.unsigned,
                            request.data,
                            Mask(mask),
                            None,
                        )
                        break

                await self.send_write_response()
                continue



from __future__ import annotations

import enum


class UnaryOperator(enum.Enum):
    BOOL = enum.auto()
    NEG = enum.auto()
    POS = enum.auto()
    INV = enum.auto()
    NOT = enum.auto()
    ABS = enum.auto()


class BinaryOperator(enum.Enum):
    AND = enum.auto()
    OR = enum.auto()

    BIT_AND = enum.auto()
    BIT_OR = enum.auto()
    BIT_XOR = enum.auto()

    ADD = enum.auto()
    SUB = enum.auto()

    MUL = enum.auto()
    DIV = enum.auto()
    MOD = enum.auto()
    FLOOR_DIV = enum.auto()

    CONCAT = enum.auto()

    LSHIFT = enum.auto()
    RSHIFT = enum.auto()


class ComparisonOperator(enum.Enum):
    EQ = enum.auto()
    NE = enum.auto()
    GT = enum.auto()
    LT = enum.auto()
    GE = enum.auto()
    LE = enum.auto()


class _IntrinsicOp:
    def __init__(self, result):
        self.result = result


class _IntrinsicUnaryOp(_IntrinsicOp):
    def __init__(self, op: UnaryOperator, result, arg):
        super().__init__(result)
        self.op = op
        self.arg = arg


class _IntrinsicBinOp(_IntrinsicOp):
    def __init__(self, op: BinaryOperator, result, lhs, rhs):
        super().__init__(result)
        self.op = op
        self.lhs = lhs
        self.rhs = rhs


class _IntrinsicComparison(_IntrinsicOp):
    def __init__(self, op: ComparisonOperator, result, lhs, rhs):
        super().__init__(result)
        self.op = op
        self.lhs = lhs
        self.rhs = rhs


class _IntrinsicConstElemAccess:
    def __init__(self, obj):
        # dummy wrapper type
        # only used to indicate to the parser, that
        # an item access was performed with constant arguments

        self.obj = obj


class _IntrinsicElemAccess:
    def __init__(self, obj, index, index_temp):
        # when __getitem__ is called with a non-constant argument
        # store the current state of the index in a temporary
        # and use the value when assignments to the value are performed

        self.obj = obj
        self.index = index
        self.index_temp = index_temp


class AssignMode(enum.Enum):
    VALUE = enum.auto()
    NEXT = enum.auto()
    PUSH = enum.auto()
    AUTO = enum.auto()
    _TEMP = enum.auto()


class _IntrinsicAssignment:
    def __init__(self, target, source, mode):
        self.target = target
        self.source = source
        self.mode = mode


class _IntrinsicDeclaration:
    def __init__(self, new_obj, assigned_value, delayed_init=False):
        self.new_obj = new_obj
        self.assigned_value = assigned_value
        self.delayed_init = delayed_init


class _IntrinsicSynthesizableFunctionCall:
    """
    when returned from an intrinsic function the callable
    is evaluated using the given args/kwargs at the location
    the intrinsic function was called.
    """

    def __init__(self, callable, args, kwargs):
        assert isinstance(args, list)
        assert isinstance(kwargs, dict)

        self.callable = callable
        self.args = args
        self.kwargs = kwargs



from __future__ import annotations

from cohdl.std.axi.axi4_light.base import Axi4Light
from cohdl.std.reg import reg32
from cohdl import Signal, Unsigned

from cohdl import Signal
from cohdl.std.axi.axi4_light.base import Axi4Light
from cohdl.std.reg import reg32

import cohdl
from cohdl import Port, Bit, BitVector, Unsigned, Null, Full
from cohdl import std
from cohdl.std.axi import axi4_light as axi

from typing import Annotated as Ann
from cohdl.std.reg import to_system_rdl


class AxiBaseEntity(cohdl.Entity):
    clk = Port.input(Bit)
    reset = Port.input(Bit)

    axi_awaddr = Port.input(Unsigned[32])
    axi_awprot = Port.input(Unsigned[3])
    axi_awvalid = Port.input(Bit)
    axi_awready = Port.output(Bit, default=Null)

    axi_wdata = Port.input(BitVector[32])
    axi_wstrb = Port.input(BitVector[4])
    axi_wvalid = Port.input(Bit)
    axi_wready = Port.output(Bit, default=Null)

    axi_bresp = Port.output(BitVector[2], default=Null)
    axi_bvalid = Port.output(Bit, default=Null)
    axi_bready = Port.input(Bit)

    axi_araddr = Port.input(Unsigned[32])
    axi_arprot = Port.input(Unsigned[3])
    axi_arvalid = Port.input(Bit)
    axi_arready = Port.output(Bit, default=Null)

    axi_rdata = Port.output(BitVector[32], default=Null)
    axi_rresp = Port.output(BitVector[2], default=Null)
    axi_rvalid = Port.output(Bit, default=Null)
    axi_rready = Port.input(Bit)

    def impl_registers(self, connection: axi.Axi4Light):
        raise NotImplementedError()

    def architecture(self):
        clk = std.Clock(self.clk)
        reset = std.Reset(self.reset)

        axi_con = axi.Axi4Light(
            clk=clk,
            reset=reset,
            wraddr=axi.Axi4Light.WrAddr(
                valid=self.axi_awvalid,
                ready=self.axi_awready,
                awaddr=self.axi_awaddr,
                awprot=self.axi_awprot,
            ),
            wrdata=axi.Axi4Light.WrData(
                valid=self.axi_wvalid,
                ready=self.axi_wready,
                wdata=self.axi_wdata,
                wstrb=self.axi_wstrb,
            ),
            wrresp=axi.Axi4Light.WrResp(
                valid=self.axi_bvalid,
                ready=self.axi_bready,
                bresp=self.axi_bresp,
            ),
            rdaddr=axi.Axi4Light.RdAddr(
                valid=self.axi_arvalid,
                ready=self.axi_arready,
                araddr=self.axi_araddr,
                arprot=self.axi_arprot,
            ),
            rddata=axi.Axi4Light.RdData(
                valid=self.axi_rvalid,
                ready=self.axi_rready,
                rdata=self.axi_rdata,
                rresp=self.axi_rresp,
            ),
        )

        self.impl_registers(axi_con)


class ClkDevice(reg32.RegFile, word_count=2):
    class Ctrl(reg32.Register):
        enable: reg32.MemField[0]
        prescaler: reg32.MemUField[31:16]

    ctrl: Ctrl[0]
    count: reg32.UWord[4]

    def _config_(self):
        self._counter = Signal[Unsigned[32]](0)

    def _impl_sequential_(self) -> None:
        if self.ctrl.enable:
            if self._counter >= self.ctrl.prescaler.val():
                self.count <<= self.count.raw + 1
                self._counter <<= 0
            else:
                self._counter <<= self._counter + 1


class AddrMap(reg32.AddrMap, word_count=64):
    clk_1: Ann[ClkDevice[0x00], "asdfasdfasdf"]
    clk_2: ClkDevice[0x08]
    clk_3: ClkDevice[0x10]
    clk_4: ClkDevice[0x18]

    def _config_(self):
        self.clk_1._config_()
        self.clk_2._config_()
        self.clk_3._config_()
        self.clk_4._config_()


class ExampleEntity(AxiBaseEntity):
    def impl_registers(self, connection: Axi4Light):
        connection.connect_addr_map(AddrMap())


print(to_system_rdl(AddrMap))



from __future__ import annotations
from ._assignable_type import AssignableType
from cohdl._core import BitVector

from cohdl import Bit, BitVector, Unsigned, Signed
from typing import overload, TypeVar

T = TypeVar("T")

#
#
#

class FieldBit:
    def __new__(cls, source: BitVector) -> Bit: ...

class FieldBitVector:
    def __new__(cls, source: BitVector) -> BitVector: ...

    BitVector: type[FieldBitVector]
    Signed: type[_FieldSigned]
    Unsigned: type[_FieldUnsigned]

class _FieldSigned(FieldBitVector):
    def __new__(cls, source: BitVector) -> Signed: ...

class _FieldUnsigned(FieldBitVector):
    def __new__(cls, source: BitVector) -> Unsigned: ...

class _MetaField(type):
    @overload
    def __getitem__(cls, arg: int) -> type[FieldBit]: ...
    @overload
    def __getitem__(cls, arg: slice) -> type[FieldBitVector]: ...
    @overload
    def __getitem__(
        cls, arg: tuple[slice, type[BitVector]]
    ) -> type[FieldBitVector]: ...
    @overload
    def __getitem__(cls, arg: tuple[slice, type[Signed]]) -> type[_FieldSigned]: ...
    @overload
    def __getitem__(cls, arg: tuple[slice, type[Unsigned]]) -> type[_FieldUnsigned]: ...

class Field(metaclass=_MetaField):
    pass

#
#
#

class _MetaBitField(type):
    def __getitem__(cls, width: int) -> type[_BitFieldInst]: ...

class _MetaBitFieldInst(type):
    @overload
    def __getitem__(cls: type[T], offset: int) -> type[T]: ...
    @overload
    def __getitem__(cls: type[T], slice: slice) -> type[T]: ...

class _BitFieldInst(BitField, metaclass=_MetaBitFieldInst):
    pass

class BitField(AssignableType, metaclass=_MetaBitField):
    _width_: int
    _offset_: int = 0
    Field: type[Field]



from __future__ import annotations

import typing
from typing import TypeVar
from ._primitive_type import _PrimitiveType

T = TypeVar("T")
Self = TypeVar("Self")

class Array(typing.Generic[T], _PrimitiveType):
    _elemtype_: type[T]
    _count_: int

    def __init__(self, val=None): ...
    def __len__(self) -> int: ...
    def __getitem__(self, slice): ...
    def __setitem__(self, slice, arg): ...
    def copy(self: Self) -> Self: ...
    def __hash__(self) -> int: ...



from __future__ import annotations

import enum

from cohdl._core._intrinsic import (
    _intrinsic,
    _intrinsic_replacement,
    _intrinsic_property,
)
from cohdl._core import _intrinsic_operations as intr_op
from cohdl._core._intrinsic_operations import AssignMode

from cohdl._core._primitive_type import is_primitive, is_primitive_type
from cohdl._core._integer import Integer
from cohdl._core._boolean import _Boolean, Null
from cohdl._core._bit_vector import BitVector
from cohdl._core._signed import Signed
from cohdl._core._unsigned import Unsigned

#
#
#


class RefSpec:
    def is_constant(self) -> bool:
        pass

    def simplify(self):
        # combine all constant values
        pass


class Offset(RefSpec):
    def __init__(self, offset, base_offset: list | None):
        self.offset = offset
        self.base_offset = base_offset if base_offset is not None else []
        self.obj = None

    def copy(self):
        new = Offset(self.offset, self.base_offset)
        new.obj = self.obj
        return new

    def simplify(self):
        base_offset = []
        const_sum = 0

        for off in self.base_offset:
            if not isinstance(off, TypeQualifier):
                assert isinstance(off, int)
                const_sum += off
            else:
                base_offset.append(off)

        if not isinstance(self.offset, TypeQualifier):
            assert isinstance(self.offset, int)
            self.offset += const_sum
            self.base_offset = base_offset
        else:
            if const_sum != 0:
                self.base_offset = [const_sum, *base_offset]
            else:
                self.base_offset = base_offset

    def is_constant(self) -> bool:
        return not isinstance(self.offset, TypeQualifier) and not any(
            isinstance(base, TypeQualifier) for base in self.base_offset
        )


class Slice(RefSpec):
    def __init__(self, start, stop, base_offset: list | None) -> None:
        self.start = start
        self.stop = stop
        self.base_offset = base_offset if base_offset is not None else []
        self.obj = None

    def copy(self):
        new = Slice(self.start, self.stop, self.base_offset)
        new.obj = self.obj
        return new

    def simplify(self):
        base_offset = []
        const_sum = 0

        for off in self.base_offset:
            if not isinstance(off, TypeQualifier):
                assert isinstance(off, int)
                const_sum += off
            else:
                base_offset.append(off)

        assert isinstance(self.start, int)
        assert isinstance(self.stop, int)

        self.start += const_sum
        self.stop += const_sum
        self.base_offset = base_offset

    def is_constant(self) -> bool:
        return (
            not isinstance(self.start, TypeQualifier)
            and not isinstance(self.stop, TypeQualifier)
            and not any(isinstance(base, TypeQualifier) for base in self.base_offset)
        )


#
#
#


class Attribute:
    name: str
    attr_type: type

    def __init__(self, value):
        assert isinstance(value, self.attr_type)
        self.value = value

    def __init_subclass__(cls, **kwargs) -> None:
        cls.name = kwargs.get("name", cls.__name__)
        cls.attr_type = kwargs["type"]


#
#
#


def _decay(val):
    if isinstance(val, TypeQualifier):
        return val._value
    return val


class _TypeQualifier(type):
    _Wrapped: type
    _SubTypes: dict[type, type]
    _Qualifier: _TypeQualifier

    @_intrinsic
    def __getitem__(cls, Wrapped: type | tuple):
        # direction only used for ports

        if isinstance(Wrapped, tuple):
            WrappedType, direction = Wrapped
        else:
            WrappedType = Wrapped
            direction = None

        if WrappedType is bool:
            WrappedType = _Boolean
        elif WrappedType is int:
            WrappedType = Integer

        type_spec = (WrappedType, direction)

        if type_spec in cls._SubTypes:
            return cls._SubTypes[type_spec]

        # issubclass(Signal[BitVector[32]], Signal[BitVector]) should return True
        if issubclass(WrappedType, BitVector):
            if hasattr(WrappedType, "_width"):
                if issubclass(WrappedType, Unsigned):
                    if direction is None:
                        parent_cls = type(
                            cls.__name__,
                            (
                                cls[Unsigned],
                                cls[BitVector[WrappedType._width]],
                            ),
                            {},
                        )
                    else:
                        parent_cls = type(
                            cls.__name__,
                            (
                                cls[Unsigned, direction],
                                cls[BitVector[WrappedType._width], direction],
                            ),
                            {},
                        )
                elif issubclass(WrappedType, Signed):
                    if direction is None:
                        parent_cls = type(
                            cls.__name__,
                            (
                                cls[Signed],
                                cls[BitVector[WrappedType._width]],
                            ),
                            {},
                        )
                    else:
                        print(cls, WrappedType._width, direction)
                        parent_cls = type(
                            cls.__name__,
                            (
                                cls[Signed, direction],
                                cls[BitVector[WrappedType._width], direction],
                            ),
                            {},
                        )
                else:
                    parent_cls = (
                        cls[BitVector]
                        if direction is None
                        else cls[BitVector, direction]
                    )
            else:
                if WrappedType is BitVector:
                    parent_cls = cls
                elif WrappedType is Unsigned or WrappedType is Signed:
                    if direction is None:
                        parent_cls = cls[BitVector]
                    else:
                        parent_cls = cls[BitVector, direction]
                else:
                    raise AssertionError(f"invalid generic argument to type qualifier")
        else:
            parent_cls = cls

        if issubclass(parent_cls, Port):
            assert isinstance(direction, Port.Direction), f"direction = {direction}"

            new_type = type(
                cls.__name__,
                (parent_cls, Signal[WrappedType]),
                {"_Wrapped": WrappedType, "_direction": direction},
            )
        else:
            assert direction is None, "direction can only be set on ports"
            new_type = type(cls.__name__, (parent_cls,), {"_Wrapped": WrappedType})

        cls._SubTypes[type_spec] = new_type
        return new_type

    @_intrinsic
    def __str__(cls):
        if not hasattr(cls, "_Wrapped"):
            return f"{cls.__name__}"

        if not hasattr(cls, "_direction"):
            return f"{cls.__name__}[{cls._Wrapped}]"
        return f"{cls.__name__}[{cls._Wrapped}, {cls._direction}]"

    @_intrinsic
    def __repr__(cls):
        return cls.__str__()

    @property
    def type(cls):
        return cls._Wrapped

    @property
    def qualifier(cls):
        return cls._Qualifier

    #
    #
    #


class TypeQualifierBase:
    @_intrinsic
    def decay(val):
        if isinstance(val, TypeQualifier):
            return val._value
        elif isinstance(val, TypeQualifierBase):
            return val.decay()
        return val


class TypeQualifier(TypeQualifierBase, metaclass=_TypeQualifier):
    @property
    def type(cls):
        return cls._Wrapped

    @property
    def qualifier(self):
        return self._Qualifier

    @property
    def width(self):
        return self._Wrapped.width

    _intrinsic(width.fget)

    #
    #
    #

    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_Wrapped"):
            value = args[0]
            decayed = _decay(value)
            wrapped_class = cls[type(decayed)]
        else:
            wrapped_class = cls

        if not is_primitive_type(wrapped_class._Wrapped):
            return wrapped_class._Wrapped(*args, **kwargs, _qualifier_=cls._Qualifier)
        else:
            return object.__new__(wrapped_class)

    @_intrinsic
    def __init__(
        self,
        value=None,
        *,
        name: str | None = None,
        attributes: dict | None = None,
        noreset: bool = False,
        maybe_uninitialized: bool = False,
        _root: TypeQualifier | None = None,
        _ref_spec: list[RefSpec] | None = None,
    ):
        if _root is not None:
            assert (
                _root.qualifier is self.qualifier
            ), f"internal error: _root.qualifier != self.qualifier {_root.qualifier} {self.qualifier}"
            assert type(value) is self.type, "internal error: type(value) != self.type"
            self._value = value
        else:
            value = TypeQualifier.decay(value)
            self._value = type(self)._Wrapped(value)

        self._name = name
        self._default = (
            None
            if (value is None) or (isinstance(self, Temporary))
            else type(self)._Wrapped(value)
        )
        self._attributes = [] if attributes is None else attributes
        self._noreset = noreset
        self._maybe_uninitialized = maybe_uninitialized

        self._root = self if _root is None else _root

        if _ref_spec is None:
            self._ref_spec = []
        else:
            self._ref_spec = list(_ref_spec)

            if len(self._ref_spec) != 0:
                self._ref_spec[-1] = _ref_spec[-1].copy()
                self._ref_spec[-1].obj = self

    @_intrinsic
    def name(self):
        return self._name

    @_intrinsic
    def set_name(self, name: str):
        self._name = name

    @_intrinsic
    def has_default(self):
        return self._default is not None

    @_intrinsic
    def default(self):
        return self._default

    @_intrinsic
    def __hash__(self):
        # Generate a unique hash for each TypeQualifier
        # so qualified types can be used as dictionary keys.
        return id(self)

    @_intrinsic
    def __str__(self):
        return f"{type(self)}({self._value!s})"

    @_intrinsic
    def __repr__(self):
        return f"{type(self)}({self._value!s})"

    @_intrinsic
    def __len__(self):
        return self._value.__len__()

    @_intrinsic
    def __iter__(self):
        if len(self._ref_spec) != 0 and isinstance(self._ref_spec[-1], Slice):
            offset = self._ref_spec[-1].stop
            prev = self._ref_spec[:-1]
        else:
            offset = 0
            prev = self._ref_spec

        for nr, elem in enumerate(self._value):
            yield self.qualifier[type(elem)](
                elem, _ref_spec=[*prev, Offset(offset + nr, [])], _root=self._root
            )

    @_intrinsic
    def get(self):
        return self._value

    @_intrinsic
    def __getitem__(self, arg):
        if isinstance(arg, (tuple, list)):
            first, *rest = arg

            if len(rest) == 0:
                if isinstance(first, slice):
                    return self.__getitem__(first)
                else:
                    return self.__getitem__(slice(first, first))
            else:
                return self.__getitem__(first) @ self.__getitem__(rest)

        ref_spec = self._ref_spec

        if len(ref_spec) != 0 and isinstance(ref_spec[-1], Slice):
            last_ref = ref_spec[-1]
            prev = ref_spec[:-1]
            base_offset = [*last_ref.base_offset, last_ref.stop]
        else:
            prev = ref_spec
            base_offset = []

        if isinstance(arg, tuple):
            raise NotImplementedError()
        elif isinstance(arg, slice):
            assert arg.step is None, "step parameter in slice is not allowed"

            if isinstance(arg.start, int):
                assert isinstance(arg.stop, int)

                result = self._value[arg.start : arg.stop]

                return self.qualifier[type(result)](
                    result,
                    _root=self._root,
                    _ref_spec=[*prev, Slice(arg.start, arg.stop, base_offset)],
                )
            else:
                raise NotImplemented()
        else:
            if isinstance(arg, int):
                result = self._value[arg]
                return self.qualifier[type(result)](
                    result,
                    _root=self._root,
                    _ref_spec=[*prev, Offset(arg, base_offset)],
                )
            elif isinstance(arg, TypeQualifier):
                result = self._value[0]

                return self.qualifier[type(result)](
                    result,
                    _root=self._root,
                    _ref_spec=[*prev, Offset(arg, base_offset)],
                )
            else:
                raise NotImplemented()

    def lsb(self, count=None, rest=None):
        val_width = self._value.width
        if count is not None:
            if rest is not None:
                assert (
                    count + rest == val_width
                ), "the sum of count and reset does not match the value width"
            return self[count - 1 : 0]
        elif rest is not None:
            vec_count = val_width - rest
            return self[vec_count - 1 : 0]
        else:
            return self[0]

    def msb(self, count=None, rest=None):
        val_width = self._value.width
        if count is not None:
            if rest is not None:
                assert (
                    count + rest == val_width
                ), "the sum of count and reset does not match the value width"
            return self[val_width - 1 : val_width - count]
        elif rest is not None:
            vec_count = val_width - rest
            return self[val_width - 1 : val_width - vec_count]
        else:
            return self[val_width - 1]

    def right(self, count=None, rest=None):
        val_width = self._value.width
        if count is not None:
            if rest is not None:
                assert (
                    count + rest == val_width
                ), "the sum of count and reset does not match the value width"
            return self[count - 1 : 0]
        elif rest is not None:
            vec_count = val_width - rest
            return self[vec_count - 1 : 0]
        else:
            return self[0]

    def left(self, count=None, rest=None):
        val_width = self._value.width
        if count is not None:
            if rest is not None:
                assert (
                    count + rest == val_width
                ), "the sum of count and reset does not match the value width"
            return self[val_width - 1 : val_width - count]
        elif rest is not None:
            vec_count = val_width - rest
            return self[val_width - 1 : val_width - vec_count]
        else:
            return self[val_width - 1]

    def resize(self, target_width: int | None = None, *, zeros: int = 0):
        type = self.type
        padded_width = self.width + zeros

        if target_width is None:
            result_width = padded_width
        else:
            result_width = target_width

        assert (
            padded_width <= result_width
        ), "width zero extended value exceeds target width"

        if zeros == 0:
            padded = self
        else:
            padded = self @ BitVector[zeros](Null)

        if issubclass(type, Signed):
            return Temporary[Signed[result_width]](padded.signed)
        elif issubclass(type, Unsigned):
            return Temporary[Unsigned[result_width]](padded.unsigned)
        else:
            raise AssertionError(
                "resize is only defined for Signed and Unsigned arguments"
            )

    def copy(self):
        return Temporary(self)

    @_intrinsic
    def _assign_(self, value, assign_mode: AssignMode):
        if assign_mode is AssignMode.NEXT:
            assert isinstance(
                self, Signal
            ), "only Signal objects can use next assignment"
            self <<= value
        elif assign_mode is AssignMode.PUSH:
            assert isinstance(
                self, Signal
            ), "only Signal objects can use push assignment"
            self ^= value
        elif assign_mode is AssignMode.VALUE:
            assert isinstance(
                self, Variable
            ), "only Variable objects can use value assignment"
            self @= value
        else:
            raise AssertionError(f"invalid assign_mode {assign_mode}")

    #
    #
    #

    @_intrinsic
    def __bool__(self):
        return bool(self._value)

    @_intrinsic
    def __and__(self, other):
        result = self._value & _decay(other)
        return Temporary[type(result)](result)

    @_intrinsic
    def __rand__(self, other):
        result = _decay(other).__and__(self._value)
        return Temporary[type(result)](result)

    @_intrinsic
    def __or__(self, other):
        result = self._value.__or__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __ror__(self, other):
        result = _decay(other).__or__(self._value)
        return Temporary[type(result)](result)

    @_intrinsic
    def __xor__(self, other):
        result = self._value.__xor__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __rxor__(self, other):
        result = _decay(other).__xor__(self._value)
        return Temporary[type(result)](result)

    @_intrinsic
    def __matmul__(self, other):
        result = self._value.__matmul__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __rmatmul__(self, other):
        result = _decay(other).__matmul__(self._value)
        return Temporary[type(result)](result)

    #
    # numeric operators
    #

    @_intrinsic
    def __add__(self, other):
        result = self._value.__add__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __radd__(self, other):
        result = self._value.__radd__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __sub__(self, other):
        result = self._value.__sub__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __rsub__(self, other):
        result = self._value.__rsub__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __mul__(self, other):
        result = self._value.__mul__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __rmul__(self, other):
        result = self._value.__rmul__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __floordiv__(self, other):
        result = self._value.__floordiv__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __rfloordiv__(self, other):
        result = self._value.__rfloordiv__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __mod__(self, other):
        result = self._value.__mod__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __rmod__(self, other):
        result = self._value.__rmod__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __lshift__(self, rhs):
        result = self._value.__lshift__(_decay(rhs))
        return Temporary[type(result)](result)

    @_intrinsic
    def __rshift__(self, rhs):
        result = self._value.__rshift__(_decay(rhs))
        return Temporary[type(result)](result)

    #
    # compare
    #

    @_intrinsic
    def __eq__(self, other):
        result = self._value.__eq__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __ne__(self, other):
        result = self._value.__ne__(_decay(other))
        return Temporary[type(result)](result)

    @_intrinsic
    def __lt__(self, other):
        result = self._value.__lt__(_decay(other))
        return result if result is NotImplemented else Temporary[type(result)](result)

    @_intrinsic
    def __gt__(self, other):
        result = self._value.__gt__(_decay(other))
        return result if result is NotImplemented else Temporary[type(result)](result)

    @_intrinsic
    def __le__(self, other):
        result = self._value.__le__(_decay(other))
        return result if result is NotImplemented else Temporary[type(result)](result)

    @_intrinsic
    def __ge__(self, other):
        result = self._value.__ge__(_decay(other))
        return result if result is NotImplemented else Temporary[type(result)](result)

    #
    # unary operators
    #

    @_intrinsic
    def __abs__(self):
        result = self._value.__abs__()
        return Temporary[type(result)](result)

    @_intrinsic
    def __inv__(self):
        result = self._value.__inv__()
        return Temporary[type(result)](result)

    @_intrinsic
    def __neg__(self):
        result = self._value.__neg__()
        return Temporary[type(result)](result)

    @_intrinsic
    def __pos__(self):
        result = self._value.__pos__()
        return Temporary[type(result)](result)

    #
    #
    # replacements
    #
    #

    @_intrinsic_replacement(__init__)
    def _init_replacement(
        self,
        value=None,
        *,
        name: str | None = None,
        attributes: dict | None = None,
        maybe_uninitialized: bool = False,
        _root: TypeQualifier | None = None,
        _ref_spec: list[RefSpec] | None = None,
    ):
        assert _root is None, "internal error: _root is None"
        assert _ref_spec is None, "internal error: _ref_spec is None"
        # set default to None, because locally defined
        # Signals/Variables cannot be used before they are constructed
        # and thus initialized
        self.__init__(
            None,
            name=name,
            attributes=attributes,
            maybe_uninitialized=maybe_uninitialized,
        )

        if value is None or is_primitive(value) and value._is_uninitialized():
            return intr_op._IntrinsicDeclaration(self, None)
        return intr_op._IntrinsicDeclaration(self, value)

    @_intrinsic_replacement(__bool__)
    def _bool_replacement(self):
        return intr_op._IntrinsicUnaryOp(
            intr_op.UnaryOperator.BOOL, Temporary[bool](self.__bool__()), self
        )

    @_intrinsic_replacement(__or__)
    def _or_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.BIT_OR, self.__or__(other), self, other
        )

    @_intrinsic_replacement(__ror__)
    def _ror_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.BIT_OR, self.__ror__(other), other, self
        )

    @_intrinsic_replacement(__and__)
    def _and_replacement_(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.BIT_AND, self.__and__(other), self, other
        )

    @_intrinsic_replacement(__rand__)
    def _rand_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.BIT_AND, self.__rand__(other), other, self
        )

    @_intrinsic_replacement(__xor__)
    def _and_replacement_(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.BIT_XOR, self.__xor__(other), self, other
        )

    @_intrinsic_replacement(__rxor__)
    def _rand_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.BIT_XOR, self.__rxor__(other), other, self
        )

    @_intrinsic_replacement(__matmul__)
    def _matmul_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.CONCAT, self.__matmul__(other), self, other
        )

    @_intrinsic_replacement(__rmatmul__)
    def _rmatmul_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.CONCAT, self.__rmatmul__(other), other, self
        )

    @_intrinsic_replacement(__getitem__)
    def __getitem_replacement(self, arg):
        if isinstance(arg, (tuple, list)):

            def synthesizable_impl(s, arg):
                first, *rest = arg

                if len(rest) == 0:
                    if isinstance(first, slice):
                        return self.__getitem__(first)
                    else:
                        return self.__getitem__(slice(first, first))
                else:
                    return self.__getitem__(first) @ synthesizable_impl(s, rest)

            return intr_op._IntrinsicSynthesizableFunctionCall(
                synthesizable_impl, [self, arg], {}
            )
        elif isinstance(arg, slice):
            assert arg.step is None, "step argument not allowed in slice"

            if isinstance(arg.start, int):
                assert isinstance(arg.stop, int)
                return intr_op._IntrinsicConstElemAccess(self.__getitem__(arg))
            else:
                raise NotImplemented()
        else:
            if isinstance(arg, int):
                return intr_op._IntrinsicConstElemAccess(self.__getitem__(arg))
            elif isinstance(arg, TypeQualifier):
                index = arg
                index_temp = Temporary[index.type]()
                result = self.__getitem__(index_temp)
                return intr_op._IntrinsicElemAccess(result, index, index_temp)
            else:
                raise NotImplemented()

    @_intrinsic_replacement(_assign_, assignment_spec=(0, 1))
    def _assign_replacement(self, value, assign_mode: AssignMode):
        if assign_mode is AssignMode.AUTO:
            if isinstance(self, Signal):
                assign_mode = AssignMode.NEXT
            else:
                assert isinstance(
                    self, Variable
                ), "AssignMode.AUTO only valid for signals and variables"
                assign_mode = AssignMode.VALUE

        if assign_mode is AssignMode.NEXT:
            assert isinstance(
                self, Signal
            ), "only Signal objects can use next assignment"
            return self._next_setter_replacement(value)
        elif assign_mode is AssignMode.PUSH:
            assert isinstance(
                self, Signal
            ), "only Signal objects can use push assignment"
            return self._push_setter_replacement(value)
        elif assign_mode is AssignMode.VALUE:
            assert isinstance(
                self, Variable
            ), "only Variable objects can use value assignment"
            return self._value_setter_replacement(value)

        raise AssertionError(f"invalid assign_mode {assign_mode}")

    #
    # numeric
    #

    @_intrinsic_replacement(__add__)
    def _add_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.ADD, self.__add__(other), self, other
        )

    @_intrinsic_replacement(__radd__)
    def _radd_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.ADD, self.__radd__(other), other, self
        )

    @_intrinsic_replacement(__sub__)
    def _sub_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.SUB, self.__sub__(other), self, other
        )

    @_intrinsic_replacement(__rsub__)
    def _rsub_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.SUB, self.__rsub__(other), other, self
        )

    @_intrinsic_replacement(__mul__)
    def _mul_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.MUL, self.__mul__(other), self, other
        )

    @_intrinsic_replacement(__rmul__)
    def _rmul_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.MUL, self.__rmul__(other), other, self
        )

    @_intrinsic_replacement(__floordiv__)
    def _floordiv_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.FLOOR_DIV, self.__floordiv__(other), self, other
        )

    @_intrinsic_replacement(__rfloordiv__)
    def _rfloordiv_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.FLOOR_DIV, self.__rfloordiv__(other), other, self
        )

    @_intrinsic_replacement(__mod__)
    def _mod_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.MOD, self.__mod__(other), self, other
        )

    @_intrinsic_replacement(__rmod__)
    def _rmod_replacement(self, other):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.MOD, self.__rmod__(other), other, self
        )

    @_intrinsic_replacement(__lshift__)
    def _lshift_replacement(self, rhs):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.LSHIFT, self.__lshift__(rhs), self, rhs
        )

    @_intrinsic_replacement(__rshift__)
    def _rshift_replacement(self, rhs):
        return intr_op._IntrinsicBinOp(
            intr_op.BinaryOperator.RSHIFT, self.__rshift__(rhs), self, rhs
        )

    #
    # compare
    #

    @_intrinsic_replacement(__eq__)
    def _eq_replacement(self, other):
        result = self.__eq__(other)

        return (
            result
            if result is NotImplemented
            else intr_op._IntrinsicComparison(
                intr_op.ComparisonOperator.EQ, result, self, other
            )
        )

    @_intrinsic_replacement(__ne__)
    def _ne_replacement(self, other):
        result = self.__ne__(other)

        return (
            result
            if result is NotImplemented
            else intr_op._IntrinsicComparison(
                intr_op.ComparisonOperator.NE, result, self, other
            )
        )

    @_intrinsic_replacement(__lt__)
    def _lt_replacement(self, other):
        result = self.__lt__(other)

        return (
            result
            if result is NotImplemented
            else intr_op._IntrinsicComparison(
                intr_op.ComparisonOperator.LT, result, self, other
            )
        )

    @_intrinsic_replacement(__gt__)
    def _gt_replacement(self, other):
        result = self.__gt__(other)

        return (
            result
            if result is NotImplemented
            else intr_op._IntrinsicComparison(
                intr_op.ComparisonOperator.GT, result, self, other
            )
        )

    @_intrinsic_replacement(__le__)
    def _le_replacement(self, other):
        result = self.__le__(other)

        return (
            result
            if result is NotImplemented
            else intr_op._IntrinsicComparison(
                intr_op.ComparisonOperator.LE, result, self, other
            )
        )

    @_intrinsic_replacement(__ge__)
    def _ge_replacement(self, other):
        result = self.__ge__(other)

        return (
            result
            if result is NotImplemented
            else intr_op._IntrinsicComparison(
                intr_op.ComparisonOperator.GE, result, self, other
            )
        )

    #
    # unary operators
    #

    @_intrinsic_replacement(__abs__)
    def _abs_replacement(self):
        result = self.__abs__()
        return intr_op._IntrinsicUnaryOp(intr_op.UnaryOperator.ABS, result, self)

    @_intrinsic_replacement(__inv__)
    def _inv_replacement(self):
        result = self.__inv__()
        return intr_op._IntrinsicUnaryOp(intr_op.UnaryOperator.INV, result, self)

    @_intrinsic_replacement(__neg__)
    def _neg_replacement(self):
        result = self.__neg__()
        return intr_op._IntrinsicUnaryOp(intr_op.UnaryOperator.NEG, result, self)

    @_intrinsic_replacement(__pos__)
    def _pos_replacement(self):
        result = self.__pos__()
        return intr_op._IntrinsicUnaryOp(intr_op.UnaryOperator.POS, result, self)

    #
    # casts
    #

    @property
    def unsigned(self):
        cast = self._value.unsigned
        return type(self)._Qualifier[type(cast)](
            cast, _ref_spec=self._ref_spec, _root=self._root
        )

    @unsigned.setter
    def unsigned(self, value):
        self._value.unsigned = _decay(value)

    _intrinsic(unsigned.fget)
    _intrinsic(unsigned.fset)

    @property
    def signed(self):
        cast = self._value.signed
        return type(self)._Qualifier[type(cast)](
            cast, _ref_spec=self._ref_spec, _root=self._root
        )

    @signed.setter
    def signed(self, value):
        self._value.signed = _decay(value)

    _intrinsic(signed.fget)
    _intrinsic(signed.fset)

    @property
    def bitvector(self):
        cast = self._value.bitvector
        return type(self)._Qualifier[type(cast)](
            cast, _ref_spec=self._ref_spec, _root=self._root
        )

    @bitvector.setter
    def bitvector(self, value):
        self._value.bitvector = _decay(value)

    _intrinsic(bitvector.fget)
    _intrinsic(bitvector.fset)


class Signal(TypeQualifier):
    _SubTypes = {}

    @_intrinsic
    def __init__(
        self,
        value=None,
        *,
        name: str | None = None,
        attributes: dict | None = None,
        delayed_init: bool = None,
        noreset: bool = False,
        maybe_uninitialized: bool = False,
        _root: TypeQualifier | None = None,
        _ref_spec: list[RefSpec] | None = None,
    ):
        assert (
            delayed_init is None
        ), "delayed_init can only be used in synthesizable contexts"

        super().__init__(
            value,
            name=name,
            attributes=attributes,
            noreset=noreset,
            maybe_uninitialized=maybe_uninitialized,
            _root=_root,
            _ref_spec=_ref_spec,
        )

    @_intrinsic_replacement(__init__)
    def _init_replacement(
        self,
        value=None,
        *,
        name: str | None = None,
        attributes: dict | None = None,
        delayed_init: bool = False,
        noreset=None,
        maybe_uninitialized: bool = False,
        _root: TypeQualifier | None = None,
        _ref_spec: list[RefSpec] | None = None,
    ):
        assert noreset is None, "noreset is not valid in synthesizable contexts"

        # copy of TypeQualifier._init_replacement
        # with added support for delayed_init

        assert _root is None, "internal error: _root is None"
        assert _ref_spec is None, "internal error: _ref_spec is None"
        # set default to None, because locally defined
        # Signals/Variables cannot be used before they are constructed
        # and thus initialized
        self.__init__(
            None,
            name=name,
            attributes=attributes,
            maybe_uninitialized=maybe_uninitialized,
        )

        if value is None or is_primitive(value) and value._is_uninitialized():
            return intr_op._IntrinsicDeclaration(self, None, delayed_init)
        return intr_op._IntrinsicDeclaration(self, value, delayed_init)

    #
    # next assignment
    #

    @property
    def next(self):
        raise AssertionError("Signal.next can only be written")

    @next.setter
    def next(self, value):
        self._value._assign(_decay(value))

    _intrinsic(next.fget)
    _intrinsic(next.fset)

    @_intrinsic_replacement(next.fset, assignment_spec=(0, 1))
    def _next_setter_replacement(self, value):
        # assign value to check whether operation is allowed
        inp_value = _decay(value)
        self._value._assign(inp_value)
        if is_primitive(inp_value):
            return intr_op._IntrinsicAssignment(self, value, AssignMode.NEXT)
        return intr_op._IntrinsicAssignment(self, self._value.copy(), AssignMode.NEXT)

    @_intrinsic
    def __ilshift__(self, value):
        self.next = value
        return self

    _intrinsic_replacement(__ilshift__, assignment_spec=(0, 1))(
        _next_setter_replacement
    )

    #
    # push assignment
    #

    @property
    def push(self):
        raise AssertionError("Signal.push can only be written")

    @push.setter
    def push(self, value):
        assert self._default is not None, "pushed signal requires default value"
        self._value._assign(_decay(value))

    _intrinsic(push.fget)
    _intrinsic(push.fset)

    @_intrinsic_replacement(push.fset, assignment_spec=(0, 1))
    def _push_setter_replacement(self, value):
        # assign value to check wheather operation is allowed
        assert (
            self._default is not None
        ), f"pushed signal requires default value (name hint='{self._name}')"
        inp_value = _decay(value)
        self._value._assign(inp_value)
        if is_primitive(inp_value):
            return intr_op._IntrinsicAssignment(self, value, AssignMode.PUSH)
        return intr_op._IntrinsicAssignment(self, self._value.copy(), AssignMode.PUSH)

    @_intrinsic
    def __ixor__(self, value):
        self.push = value
        return self

    _intrinsic_replacement(__ixor__, assignment_spec=(0, 1))(_push_setter_replacement)


class Port(Signal):
    _SubTypes = {}
    _direction: Direction

    class Direction(enum.Enum):
        INPUT = enum.auto()
        OUTPUT = enum.auto()
        INOUT = enum.auto()

        @_intrinsic
        def __str__(self):
            match self:
                case Port.Direction.INPUT:
                    return "INPUT"
                case Port.Direction.OUTPUT:
                    return "OUTPUT"
                case Port.Direction.INOUT:
                    return "INOUT"

        @_intrinsic
        def is_input(self):
            return self is Port.Direction.INPUT

        @_intrinsic
        def is_output(self):
            return self is Port.Direction.OUTPUT

        @_intrinsic
        def is_inout(self):
            return self is Port.Direction.INOUT

    @classmethod
    @_intrinsic
    def is_input(cls):
        return cls._direction is Port.Direction.INPUT

    @classmethod
    @_intrinsic
    def is_output(cls):
        return cls._direction is Port.Direction.OUTPUT

    @classmethod
    @_intrinsic
    def is_inout(cls):
        return cls._direction is Port.Direction.INOUT

    @staticmethod
    def input(Wrapped, *, name: str | None = None) -> Port:
        return Port[Wrapped, Port.Direction.INPUT](name=name)

    @staticmethod
    def output(
        Wrapped, *, default=None, name: str | None = None, noreset=False
    ) -> Port:
        return Port[Wrapped, Port.Direction.OUTPUT](default, name=name, noreset=noreset)

    @staticmethod
    def inout(Wrapped, *, default=None, name: str | None = None, noreset=False) -> Port:
        return Port[Wrapped, Port.Direction.INOUT](default, name=name, noreset=noreset)

    @classmethod
    @_intrinsic
    def direction(cls):
        return cls._direction


class Variable(TypeQualifier):
    _SubTypes = {}

    #
    # value assignment
    #

    @property
    def value(self):
        raise AssertionError("Variable.value can only be written")

    @value.setter
    def value(self, value):
        self._value._assign(_decay(value))

    _intrinsic_property(value)

    @_intrinsic_replacement(value.fset, assignment_spec=(0, 1))
    def _value_setter_replacement(self, value):
        # assign value to check wheather operation is allowed
        inp_value = _decay(value)
        self._value._assign(inp_value)
        if is_primitive(inp_value):
            return intr_op._IntrinsicAssignment(self, value, AssignMode.VALUE)
        return intr_op._IntrinsicAssignment(self, self._value.copy(), AssignMode.VALUE)

    @_intrinsic
    def __imatmul__(self, value):
        self.next = value
        return self

    _intrinsic_replacement(__imatmul__, assignment_spec=(0, 1))(
        _value_setter_replacement
    )


class Temporary(TypeQualifier):
    _SubTypes = {}


class Generic: ...


Signal._Qualifier = Signal
Variable._Qualifier = Variable
Temporary._Qualifier = Temporary



from __future__ import annotations

class _Prefix:
    def subprefix(self, subname: str) -> _Prefix:
        """
        Create a new prefix by concatenating self with `subname`.
        """

    def name(self, name: str | None) -> str:
        """
        Returns a new name by concatenating the prefix and the given name.
        """

    def prefix_str(self) -> str:
        """
        Returns the current prefix as a string.
        """

    def __enter__(self) -> _Prefix:
        """
        Starts a prefixed context.

        All calls to `std.prefix` and `std.name` will be prefixed
        with `self.prefix_str`.
        """

    def __exit__(self, type, value, traceback) -> None: ...

def prefix(base_name: str) -> _Prefix:
    """
    Request a new unique prefix built from `base_name`.

    The `std.prefix` utility exists to make it easier to track cohdl Signals/Variables
    in the generated HDL representation. Usage of this function is optional and has no
    runtime effect.

    >>> class Uart:
    >>>     def __init__(self, flow_control=False):
    >>>         p = std.prefix("uart")
    >>>         self.rx = Signal[Bit](name=p.name("rx"))
    >>>         self.tx = Signal[Bit](name=p.name("tx"))
    >>>
    >>>         # prefix can also be used in a with statement
    >>>         # and is then applied to all nested prefixes
    >>>         # and `std.name`
    >>>         with p:
    >>>             if flow_control:
    >>>                 self.rts = Signal[Bit](name=std.name("rts"))
    >>>                 self.cts = Signal[Bit](name=std.name("cts"))
    >>>         ...
    >>>
    >>> uart1 = Uart()
    >>> # will create two signals "uart_rx" and "uart_tx"
    >>> uart2 = Uart(True)
    >>> # will create four signals "uart_1_rx", "uart_1_tx", "uart_1_rts" and "uart_1_cts"
    >>> # the `1` is added to create a unique prefix to indicate, that all these signals
    >>> # belong to the same class.
    """

def name(name: str) -> str:
    """
    When this function is called in a prefixed context (see `std.prefix`) it returns
    a new name built by concatenating the prefix with `name`.
    Outside of prefixed contexts the argument is returned unchanged.
    """

class _NamedQualifier:
    def __getitem__(self, qualifier, name: str):
        return qualifier

NamedQualifier = _NamedQualifier()
"""
`NamedQualifier` has no runtime effect. It only affects the names
of Signals/Variables/Temporaries and is used to make the generated HDL more readable.

In the following example `plain_obj` and `named_obj` are functionally identical.
The names of new Signals constructed during the initialization of `plain_obj` are
random (unless explicitly defined).
The names of new Signals constructed during the initialization of `named_obj` are
derived from the given base name `my_name`.

>>> # plain_obj and named_obj are functionally identical
>>> # the Signals created during the initialization of named_obj
>>> # are named using the prefix 'my_name'.
>>> plain_obj = cohdl.Signal(old_obj)
>>> named_obj = NamedQualifier[cohdl.Signal, "my_name"](old_obj)
>>>
"""



from __future__ import annotations

from ._bit_vector import BitVector, BitOrder
from ._bit import Bit

from ._intrinsic import _intrinsic
from ._integer import Integer

from ._boolean import Null

from cohdl.utility import Span


class Unsigned(BitVector):
    _SubTypes = {}

    @staticmethod
    @_intrinsic
    def upto(max_value: int):
        width = max_value.bit_length()
        return Unsigned[width]

    @staticmethod
    def _uint_to_binary(width: int, number: int) -> str:
        return "".join([str(int(bool(number & 2**i))) for i in range(width)][::-1])

    @staticmethod
    @_intrinsic
    def from_int(
        value: int | Integer,
        max: int | None = None,
        order: BitOrder = BitOrder.DOWNTO,
    ):
        if isinstance(value, Integer):
            value = int(value)

        if max is None:
            max = value

        assert 0 <= value <= max, f"value {value} outside allowed range ({0}-{max})"

        width = len(bin(max)) - 2

        return Unsigned[width](value)

    @staticmethod
    def from_str(default: str, order: BitOrder = BitOrder.DOWNTO) -> Unsigned:
        return Unsigned(len(default), default, order)

    @_intrinsic
    def __init__(
        self,
        val: None | BitVector | str | int = None,
    ):
        if isinstance(val, Integer):
            val = int(val)

        if isinstance(val, int):
            assert (
                0 <= val < 2**self.width
            ), f"value {val} outside allowed range ({0}-{2**self.width-1})"
            val = Unsigned._uint_to_binary(self.width, val)

        super().__init__(val)

    @classmethod
    @_intrinsic
    def min_int(cls) -> int:
        # only implemented for symmetry with Signed
        return 0

    @classmethod
    @_intrinsic
    def max_int(cls) -> int:
        return 2**cls.width - 1

    @classmethod
    @_intrinsic
    def min(cls) -> Unsigned:
        # only implemented for symmetry with Signed
        return cls(cls.min_int())

    @classmethod
    @_intrinsic
    def max(cls) -> Unsigned:
        return cls(cls.max_int())

    def _assign(self, other):
        import cohdl

        assert not isinstance(
            other, cohdl.Signed
        ), "Signed value cannot be assigned to Unsigned without explicit cast"

        if isinstance(other, Integer):
            other = int(other)

        if isinstance(other, int):
            assert 0 <= other <= self.max()

            binary = Span([*bin(other)[2:]][::-1])

            self._value.apply_zip(
                lambda bit, char: bit._assign(char), binary.iter_extend("0")
            )
        elif isinstance(other, Unsigned):
            assert (
                other.width <= self.width
            ), f"target width {self.width} is less than source width {other.width}"

            self._value.apply_zip(
                lambda bit, other_bit: bit._assign(other_bit),
                other._value.iter_extend(Bit(0)),
            )

        elif (
            isinstance(other, (BitVector, str))
            or other is cohdl.Null
            or other is cohdl.Full
        ):
            super()._assign(other)
        else:
            raise AssertionError()

    def __index__(self):
        return self.to_int()

    @_intrinsic
    def to_int(self) -> int:
        result = 0

        for shift, bit in enumerate(self._value):
            if bit:
                result |= 1 << shift

        return result

    @_intrinsic
    def add(self, rhs: Unsigned | int | Integer, target_width=None) -> Unsigned:
        if isinstance(rhs, (int, Integer)):
            rhs = rhs % 2**self.width
            rhs = Unsigned.from_int(rhs, order=self._order)

            if target_width is None:
                # adding integer, that cannot be represented
                # without explicit specification of target_width
                # is probably an error
                assert (
                    rhs.width <= self.width
                ), "added integer is not in the representable target range"
                target_width = self.width
        else:
            assert self.order == rhs.order, "bitorder missmatch"
            assert isinstance(rhs, Unsigned), "expected unsigned argument"

            if target_width is None:
                target_width = max(self.width, rhs.width)

        target = Unsigned[target_width](Null)

        carry = False

        null_bit = Bit(0)

        for a, b, t in zip(
            self._value.iter_extend(null_bit),
            rhs._value.iter_extend(null_bit),
            target._value,
        ):
            a = bool(a)
            b = bool(b)

            t._assign((a + b + carry) & 1)
            carry = (a + b + carry) > 1

        return target

    @_intrinsic
    def sub(self, rhs: Unsigned | int | Integer, target_width=None) -> Unsigned:
        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        if isinstance(rhs, int):
            rhs = -(rhs % 2**self.width)

        else:
            rhs = -rhs

        return self.add(rhs, target_width)

    @_intrinsic
    def __add__(self, rhs: Unsigned | int | Integer) -> Unsigned:
        return self.add(rhs)

    @_intrinsic
    def __radd__(self, lhs: int | Integer) -> Unsigned:
        return self.add(lhs)

    @_intrinsic
    def __sub__(self, rhs: Unsigned | int | Integer) -> Unsigned:
        return self.sub(rhs)

    @_intrinsic
    def __rsub__(self, lhs: int | Integer) -> Unsigned:
        return lhs + -self

    @_intrinsic
    def __mul__(self, rhs: Unsigned | int | Integer) -> Unsigned:
        result_width = self.width

        if isinstance(rhs, Unsigned):
            result_width = self.width + rhs.width
            rhs = rhs.to_int()
        elif isinstance(rhs, Integer):
            rhs = rhs.get_value()

        lhs = self.to_int()

        return Unsigned[result_width](lhs * rhs)

    @_intrinsic
    def __floordiv__(self, rhs: Unsigned | int | Integer) -> Unsigned:
        result_width = self.width

        if isinstance(rhs, Unsigned):
            rhs = rhs.to_int()
        elif isinstance(rhs, Integer):
            rhs = rhs.get_value()

        lhs = self.to_int()

        if rhs == 0:
            return Unsigned[result_width]()
        return Unsigned[result_width](lhs // rhs)

    @_intrinsic
    def __mod__(self, rhs: Unsigned | int | Integer) -> Unsigned:
        result_width = self.width

        if isinstance(rhs, Unsigned):
            rhs = rhs.to_int()
        elif isinstance(rhs, Integer):
            rhs = rhs.get_value()

        lhs = self.to_int()

        if rhs == 0:
            return Unsigned[result_width]()
        return Unsigned[result_width](lhs % rhs)

    @_intrinsic
    def __lshift__(self, rhs: Unsigned | int | Integer) -> Unsigned:
        i = self.to_int()
        s = int(rhs)
        val = (self.to_int() << int(rhs)) & ((1 << self.width) - 1)
        return Unsigned[self.width](val)

    @_intrinsic
    def __rshift__(self, rhs: Unsigned | int | Integer) -> Unsigned:
        val = self.to_int() >> int(rhs)
        return Unsigned[self.width](val)

    @_intrinsic
    def __neg__(self) -> Unsigned:
        return ~self + 1

    @_intrinsic
    def __eq__(self, other: BitVector | int | Integer) -> bool:
        if isinstance(other, Integer):
            other = other.get_value()

        if isinstance(other, int):
            return bool(other == self.to_int())
        return super().__eq__(other)

    @_intrinsic
    def __ne__(self, other: BitVector | int | Integer) -> bool:
        if isinstance(other, Integer):
            other = other.get_value()

        if isinstance(other, int):
            return bool(other != self.to_int())
        return super().__ne__(other)

    @_intrinsic
    def __lt__(self, rhs: Unsigned | int | Integer) -> bool:
        assert isinstance(rhs, (Unsigned, int, Integer)), f"invalid argument '{rhs}'"
        if isinstance(rhs, Unsigned):
            rhs = rhs.to_int()

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        return bool(self.to_int() < rhs)

    @_intrinsic
    def __gt__(self, rhs: Unsigned | int | Integer) -> bool:
        assert isinstance(rhs, (Unsigned, int, Integer)), f"invalid argument '{rhs}'"
        if isinstance(rhs, Unsigned):
            rhs = rhs.to_int()

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        return bool(self.to_int() > rhs)

    @_intrinsic
    def __le__(self, rhs: Unsigned | int | Integer) -> bool:
        assert isinstance(rhs, (Unsigned, int, Integer)), f"invalid argument '{rhs}'"
        if isinstance(rhs, Unsigned):
            rhs = rhs.to_int()

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        return bool(self.to_int() <= rhs)

    @_intrinsic
    def __ge__(self, rhs: Unsigned | int | Integer) -> bool:
        assert isinstance(rhs, (Unsigned, int, Integer)), f"invalid argument '{rhs}'"
        if isinstance(rhs, Unsigned):
            rhs = rhs.to_int()

        if isinstance(rhs, Integer):
            rhs = rhs.get_value()

        return bool(self.to_int() >= rhs)

    @_intrinsic
    def resize(self, target_width: int | None = None, *, zeros: int = 0):
        if target_width is None:
            target_width = self.width + zeros

        assert (
            self.width + zeros <= target_width
        ), f"width of zero extended value ({self.width}+{zeros}) exceeds target width ({target_width})"
        val = self.to_int() * 2**zeros
        return Unsigned[target_width](val)

    @_intrinsic
    def __repr__(self):
        return f"Unsigned[{self.width-1}:0]({self._bit_str()})"

    @_intrinsic
    def __str__(self):
        return f"{self.to_int()}u"



from __future__ import annotations

import cohdl
from cohdl import Bit, Port
from cohdl import std


class CondWithSideEffect:
    def __init__(self, cond, comment):
        self.cond = cond
        self.comment = comment

    def __bool__(self):
        std.comment(self.comment)
        return bool(self.cond)


class test_while_return_06(cohdl.Entity):
    clk = Port.input(Bit)
    reset = Port.input(Bit)

    def architecture(self):
        @std.sequential
        async def proc():
            while CondWithSideEffect(self.reset, "XXXXXXXXX"):
                await cohdl.true
                continue


print(std.VhdlCompiler.to_string(test_while_return_06))
